# -*- coding: utf-8 -*-
"""."""
# %% Import
# Standard library imports
import re

# Third party imports
from spyder.config.manager import CONF

# Local imports
from spyder_okvim.executor.executor_base import (
    FUNC_INFO,
    RETURN_EXECUTOR_METHOD_INFO,
    ExecutorBase,
)
from spyder_okvim.executor.executor_easymotion import ExecutorEasymotion
from spyder_okvim.executor.executor_sub import (
    ExecutorSearch,
    ExecutorSubCmd_f_t,
    ExecutorSubCmd_g,
    ExecutorSubCmd_r,
    ExecutorSubCmd_register,
    ExecutorSubCmdSneak,
    ExecutorSubMotion_a,
    ExecutorSubMotion_i,
)
from spyder_okvim.executor.executor_surround import ExecutorAddSurround
from spyder_okvim.spyder.config import CONF_SECTION


class ExecutorVisualCmd(ExecutorBase):
    """Executor for visual mode."""

    def __init__(self, vim_status):
        super().__init__(vim_status)

        cmds = 'uUoiaydxscVhHjJklLMwWbBSepP^$gG~%fFtTnN/;,"r<> \b\r*#'
        self.pattern_cmd = re.compile(r"(\d*)([{}])".format(cmds))
        self.set_cursor_pos = vim_status.cursor.set_cursor_pos
        self.set_cursor_pos_in_visual = vim_status.cursor.set_cursor_pos_in_visual
        self.apply_motion_info_in_visual = (
            self.vim_status.cursor.apply_motion_info_in_visual
        )
        self.set_block_selection_in_visual = (
            self.vim_status.cursor.set_block_selection_in_visual
        )
        self.executor_sub_g = ExecutorSubCmd_g(vim_status)
        self.executor_sub_f_t = ExecutorSubCmd_f_t(vim_status)
        self.executor_sub_r = ExecutorSubCmd_r(vim_status)
        self.executor_sub_motion_i = ExecutorSubMotion_i(vim_status)
        self.executor_sub_motion_a = ExecutorSubMotion_a(vim_status)
        self.executor_sub_register = ExecutorSubCmd_register(vim_status)
        self.executor_sub_search = ExecutorSearch(vim_status)
        self.executor_sub_easymotion = ExecutorEasymotion(vim_status)
        self.executor_sub_sneak = ExecutorSubCmdSneak(vim_status)
        self.executor_sub_surround = ExecutorAddSurround(vim_status)

    def zero(self, num=1, num_str=""):
        """Go to the start of the current line."""
        motion_info = self.helper_motion.zero()

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def V(self, num=1, num_str=""):
        """Start Visual mode per line."""
        self.vim_status.to_visual_line()

    def h(self, num=1, num_str=""):
        """Move cursor to left."""
        motion_info = self.helper_motion.h(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def k(self, num=1, num_str=""):
        """Move cursor to up."""
        motion_info = self.helper_motion.k(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def H(self, num=1, num_str=""):
        """Move cursor to the top of the page."""
        motion_info = self.helper_motion.H(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def M(self, num=1, num_str=""):
        """Move cursor to the middle of the page."""
        motion_info = self.helper_motion.M(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def L(self, num=1, num_str=""):
        """Move cursor to the bottom of the page."""
        motion_info = self.helper_motion.L(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def j(self, num=1, num_str=""):
        """Move cursor to down."""
        motion_info = self.helper_motion.j(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def J(self, num=1, num_str=""):
        """Join lines, with a minimum of two lines."""
        block_no_start = self.get_block_no_start_in_selection()
        block_no_end = self.get_block_no_end_in_selection()

        if block_no_start is None:
            self.vim_status.to_normal()
            return

        if block_no_start == block_no_end:
            block_no_end += 1

        cursor_pos_start = self.get_pos_start_in_selection()
        self.helper_action.join_lines(cursor_pos_start, block_no_start, block_no_end)
        self.vim_status.to_normal()

    def l(self, num=1, num_str=""):
        """Move cursor to right."""
        motion_info = self.helper_motion.l(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def caret(self, num=1, num_str=""):
        """Move cursor to the first non-blank character of the line."""
        motion_info = self.helper_motion.caret(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def dollar(self, num=1, num_str=""):
        """Move cursor to the end of the line."""
        motion_info = self.helper_motion.dollar(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def w(self, num=1, num_str=""):
        """Move to the next word."""
        motion_info = self.helper_motion.w(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def W(self, num=1, num_str=""):
        """Move to the next WORD."""
        motion_info = self.helper_motion.W(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def b(self, num=1, num_str=""):
        """Move to the previous word."""
        motion_info = self.helper_motion.b(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def B(self, num=1, num_str=""):
        """Move to the previous WORD."""
        motion_info = self.helper_motion.B(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def e(self, num=1, num_str=""):
        """Move to the the end of word."""
        motion_info = self.helper_motion.e(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def g(self, num=1, num_str=""):
        """Start g submode."""
        executor_sub = self.executor_sub_g

        self.set_parent_info_to_submode(executor_sub, num, num_str)
        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_visual, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def G(self, num=1, num_str=""):
        """Move to the line."""
        motion_info = self.helper_motion.G(num=num, num_str=num_str)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def tilde(self, num=1, num_str=""):
        """Switch case of the character under the cursor.

        Move the cursor to the start of selection.
        """
        self.helper_action.handle_case(None, "swap")

    def percent(self, num=1, num_str=""):
        """Go to matching bracket."""
        motion_info = self.helper_motion.percent(num=num, num_str=num_str)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def f(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_visual, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def F(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_visual, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def t(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_visual, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def T(self, num=1, num_str=""):
        """Go to the next occurrence of a character."""
        executor_sub = self.executor_sub_f_t

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_visual, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def semicolon(self, num=1, num_str=""):
        """Repeat latest f, t, f, T."""
        motion_info = self.helper_motion.semicolon(num=num, num_str=num_str)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def comma(self, num=1, num_str=""):
        """Repeat latest f, t, f, T in opposite direction."""
        motion_info = self.helper_motion.comma(num=num, num_str=num_str)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def r(self, num=1, num_str=""):
        """Replace the selected text under the cursor with input."""
        executor_sub = self.executor_sub_r
        executor_sub.pos_start = self.get_pos_start_in_selection()
        executor_sub.pos_end = self.get_pos_end_in_selection()

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [
                FUNC_INFO(self.vim_status.to_normal, False),
                FUNC_INFO(lambda: self.set_cursor_pos(executor_sub.pos_start), False),
            ]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def greater(self, num=1, num_str=""):
        """Shift lines rightwards."""
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()
        self.helper_action._indent(sel_start, sel_end)
        self.vim_status.to_normal()
        self.vim_status.cursor.draw_vim_cursor()

    def less(self, num=1, num_str=""):
        """Shift lines leftwards."""
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()
        self.helper_action._unindent(sel_start, sel_end)
        self.vim_status.to_normal()
        self.vim_status.cursor.draw_vim_cursor()

    def i(self, num=1, num_str=""):
        """Select block exclusively."""
        executor_sub = self.executor_sub_motion_i

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.set_block_selection_in_visual, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def a(self, num=1, num_str=""):
        """Select block exclusively."""
        executor_sub = self.executor_sub_motion_a

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.set_block_selection_in_visual, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def quote(self, num=1, num_str=""):
        """Set the name of register."""
        executor_sub = self.executor_sub_register

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred([])

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def d(self, num=1, num_str=""):
        """Delete text."""
        sel_start, sel_end = self.helper_action.yank(None)
        self.helper_action.delete(None, is_insert=False)
        self.vim_status.to_normal()
        self.set_cursor_pos(sel_start)
        cursor = self.get_cursor()
        if cursor.atBlockEnd() and not cursor.atBlockStart():
            self.set_cursor_pos(sel_start - 1)

    def x(self, num=1, num_str=""):
        """Delete text."""
        self.d(num, num_str)

    def c(self, num=1, num_str=""):
        """Delete text and start insert."""
        sel_start, sel_end = self.helper_action.yank(None)
        self.helper_action.delete(None, is_insert=True)
        self.vim_status.to_normal()
        self.set_cursor_pos(sel_start)
        self.get_editor().setFocus()

    def s(self, num=1, num_str=""):
        """Delete text and start insert."""
        use_sneak = CONF.get(CONF_SECTION, "use_sneak")
        if use_sneak:
            executor_sub = self.executor_sub_sneak

            self.set_parent_info_to_submode(executor_sub, num, num_str)

            executor_sub.set_func_list_deferred(
                [
                    FUNC_INFO(self.apply_motion_info_in_visual, True),
                    FUNC_INFO(
                        self.helper_motion.display_another_group_after_sneak, False
                    ),
                ]
            )

            return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)
        else:
            self.c(num, num_str)

    def S(self, num=1, num_str=""):
        """Add surroundings: parentheses, brackets, quotes."""
        self.vim_status.set_message("")
        executor_sub = self.executor_sub_surround

        executor_sub.pos_start = self.get_pos_start_in_selection()
        executor_sub.pos_end = self.get_pos_end_in_selection()

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def y(self, num=1, num_str=""):
        """Yank selected text."""
        sel_start, sel_end = self.helper_action.yank(None, is_explicit=True)

        self.vim_status.to_normal()
        self.set_cursor_pos(sel_start)

    def slash(self, num=1, num_str=""):
        """Go to the next searched text."""
        self.vim_status.set_message("")
        executor_sub = self.executor_sub_search

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_visual, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, False)

    def n(self, num=1, num_str=""):
        """Go to the next searched text."""
        motion_info = self.helper_motion.n(num=num, num_str=num_str)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def N(self, num=1, num_str=""):
        """Go to the previous searched text."""
        motion_info = self.helper_motion.N(num=num, num_str=num_str)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def o(self, num=1, num_str=""):
        """Go to other end of highlighted text."""
        sel_start = self.vim_status.get_pos_start_in_selection()
        sel_end = self.vim_status.get_pos_end_in_selection()
        cur_pos = self.get_cursor().position()
        if abs(sel_start - cur_pos) < abs(sel_end - cur_pos):
            new_pos = sel_end - 1
        else:
            new_pos = sel_start
        self.set_cursor_pos(new_pos)

    def u(self, num=1, num_str=""):
        """Make txt lowercase and move cursor to the start of selection."""
        self.helper_action.handle_case(None, "lower")

    def U(self, num=1, num_str=""):
        """Make txt uppercase and move cursor to the start of selection."""
        self.helper_action.handle_case(None, "upper")

    def p(self, num=1, num_str=""):
        """Put the text from register."""
        self.helper_action.paste_in_visual(num)

    def P(self, num=1, num_str=""):
        """Put the text from register."""
        self.helper_action.paste_in_visual(num)

    def space(self, num=1, num_str=""):
        """Move cursor to right."""
        motion_info = self.helper_motion.space(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def backspace(self, num=1, num_str=""):
        """Move cursor to left."""
        motion_info = self.helper_motion.backspace(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def enter(self, num=1, num_str=""):
        """Move cursor to down."""
        motion_info = self.helper_motion.enter(num=num)

        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def run_easymotion(self, num=1, num_str=""):
        """Run easymotion."""
        executor_sub = self.executor_sub_easymotion

        self.set_parent_info_to_submode(executor_sub, num, num_str)

        executor_sub.set_func_list_deferred(
            [FUNC_INFO(self.apply_motion_info_in_visual, True)]
        )

        return RETURN_EXECUTOR_METHOD_INFO(executor_sub, True)

    def asterisk(self, num=1, num_str=""):
        """Search word under cursor forward."""
        motion_info = self.helper_motion.asterisk(num)
        self.apply_motion_info_in_visual(motion_info)

    def sharp(self, num=1, num_str=""):
        """Search word under cursor backward."""
        motion_info = self.helper_motion.sharp(num)
        self.apply_motion_info_in_visual(motion_info)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""OkVim Widget."""
# %% Import
# Standard library imports
import os.path as osp
import sys
import threading
from functools import wraps

# Third party imports
from qtpy.QtCore import QObject, Qt, QThread, Signal, Slot
from qtpy.QtGui import QKeySequence, QTextCursor
from qtpy.QtWidgets import QHBoxLayout, QLabel, QLineEdit, QWidget
from spyder.api.config.decorators import on_conf_change
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.config.manager import CONF

# Local imports
from spyder_okvim.executor import (
    ExecutorLeaderKey,
    ExecutorNormalCmd,
    ExecutorVisualCmd,
    ExecutorVlineCmd,
)
from spyder_okvim.spyder.config import CONF_SECTION, KEYCODE2STR
from spyder_okvim.utils.path_finder import PathFinder
from spyder_okvim.utils.vim_status import InputCmdInfo, KeyInfo, VimState, VimStatus

running_coverage = "coverage" in sys.modules


def coverage_resolve_trace(fn):
    """Fix missing coverage of qthread."""

    @wraps(fn)
    def wrapped(*args, **kwargs):
        if running_coverage:
            sys.settrace(threading._trace_hook)
        fn(*args, **kwargs)

    return wrapped


class SpyderOkVimPane(PluginMainWidget):
    """."""

    def __init__(self, name=None, plugin=None, parent=None):
        """."""
        super().__init__(name, plugin, parent)
        layout = QHBoxLayout()
        layout.addWidget(QLabel("Please hide this pane."))
        self.setLayout(layout)

        self.main = parent
        self.vim_cmd = VimWidget(self.main.editor, self.main)

    def get_title(self):
        """."""
        return "Okvim"

    def get_focus_widget(self):
        """."""
        return self.vim_cmd

    def setup(self):
        """."""
        pass

    def update_actions(self):
        """."""
        pass

    @on_conf_change
    def apply_plugin_settings(self, options):
        """Apply the config settings."""
        self.vim_cmd.vim_status.search.set_color()
        self.vim_cmd.vim_status.cursor.set_config_from_conf()
        self.vim_cmd.set_leader_key()


class VimShortcut(QObject):
    """Method class for vim shortcut."""

    signal_cmd = Signal(str)

    def __init__(self, main, vim_status: VimStatus):
        super().__init__()
        self.main = main
        self.vim_status = vim_status
        self.get_editor = self.vim_status.get_editor
        self.cmd_line = None

    def pg_half_up(self):
        """Scroll window upward."""
        scroll_lines = self.vim_status.get_number_of_visible_lines() // 2
        if scroll_lines < 1:
            scroll_lines = 1

        self.signal_cmd.emit(f"{scroll_lines}k")
        self.signal_cmd.emit("^")

    def pg_up(self):
        """Scroll window upward."""
        scroll_lines = self.vim_status.get_number_of_visible_lines()
        if scroll_lines < 1:
            scroll_lines = 1

        self.signal_cmd.emit(f"{scroll_lines}k")
        self.signal_cmd.emit("^")

    def pg_half_down(self):
        """Scroll window downward."""
        scroll_lines = self.vim_status.get_number_of_visible_lines() // 2
        if scroll_lines < 1:
            scroll_lines = 1

        self.signal_cmd.emit(f"{scroll_lines}j")
        self.signal_cmd.emit("^")

    def pg_down(self):
        """Scroll window downward."""
        scroll_lines = self.vim_status.get_number_of_visible_lines()
        if scroll_lines < 1:
            scroll_lines = 1

        self.signal_cmd.emit(f"{scroll_lines}j")
        self.signal_cmd.emit("^")

    def _extract_number(self):
        """Extract the number of current cursor position."""
        cursor = self.get_editor().textCursor()

        val = None
        block_start = cursor.block().position()
        block_end = block_start + cursor.block().length()
        pos_start = cursor.position()
        pos_end = pos_start

        for _pos_end in range(pos_start + 1, block_end):
            cursor.setPosition(pos_start)
            cursor.setPosition(_pos_end, QTextCursor.KeepAnchor)
            try:
                text_selected = cursor.selectedText()
                if text_selected == "-":
                    continue
                val_ = int(text_selected)
                if text_selected.isdigit() or text_selected[0] == "-":
                    val = val_
                    pos_end = _pos_end
                else:
                    break
            except ValueError:
                break

        for _pos_start in range(pos_start, block_start - 1, -1):
            cursor.setPosition(_pos_start)
            cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
            try:
                val_ = int(cursor.selectedText())
                text_selected = cursor.selectedText()
                if text_selected.isdigit() or text_selected[0] == "-":
                    val = val_
                    pos_start = _pos_start
                else:
                    break
            except ValueError:
                break

        return val, pos_start, pos_end

    def add_num(self):
        """Add to the number at the cursor."""
        val, pos_start, pos_end = self._extract_number()

        if val is not None:
            if self.vim_status.sub_mode:
                self.cmd_line.esc_pressed()
                return

            txt = self.cmd_line.text()
            num = 1 if not txt else int(txt)

            cursor = self.get_editor().textCursor()
            cursor.setPosition(pos_start)
            cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
            cursor.insertText(str(val + num))
            self.signal_cmd.emit("h")

            # Update dot cmd
            cmd_info = InputCmdInfo(str(num), "")
            self.vim_status.input_cmd.set(cmd_info)
            key_info = KeyInfo(Qt.Key_A, "", Qt.ControlModifier, 0)
            self.vim_status.update_dot_cmd(False, key_list_to_cmd_line=[key_info])

    def subtract_num(self):
        """Subtract to the number at the cursor."""
        val, pos_start, pos_end = self._extract_number()

        if val is not None:
            if self.vim_status.sub_mode:
                self.cmd_line.esc_pressed()
                return

            txt = self.cmd_line.text()
            num = 1 if not txt else int(txt)

            cursor = self.get_editor().textCursor()
            cursor.setPosition(pos_start)
            cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
            cursor.insertText(str(val - num))
            self.signal_cmd.emit("h")

            # Update dot cmd
            cmd_info = InputCmdInfo(str(num), "")
            self.vim_status.input_cmd.set(cmd_info)
            key_info = KeyInfo(Qt.Key_X, "", Qt.ControlModifier, 0)
            self.vim_status.update_dot_cmd(False, key_list_to_cmd_line=[key_info])

    def redo(self):
        """Redo [count] changes which were undone."""
        if self.vim_status.sub_mode:
            self.cmd_line.esc_pressed()
            return

        if not self.vim_status.is_normal():
            return

        editor = self.vim_status.get_editor()
        n_block_old = editor.blockCount()

        txt = self.cmd_line.text()
        num = 1 if not txt else int(txt)

        editor = self.get_editor()
        for _ in range(num):
            editor.redo()

        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            if n_block_new > n_block_old:
                self.vim_status.set_message(f"{n_block_new - n_block_old} more lines")
            else:
                self.vim_status.set_message(f"{n_block_old - n_block_new} fewer lines")
        else:
            self.vim_status.set_message(f"{num} changes")

        self.cmd_line.esc_pressed()

    def open_path_finder(self):
        """Open path finder."""
        root_folder = self.main.projects.get_active_project_path()

        dlg = PathFinder(root_folder, self.main)
        dlg.exec_()
        path = dlg.get_path_selected()

        if osp.isfile(path):
            self.main.open_file(path)
            self.vim_status.set_focus_to_vim()

    def clear_tip_search(self):
        """Clear tooltip, search highlight."""
        self.get_editor().hide_tooltip()
        self.vim_status.cursor.set_extra_selections("vim_search", [])

class VimStateLabel(QLabel):
    """Display state of vim."""

    def __init__(self, parent):
        super().__init__(parent)
        self.change_state(VimState.INSERT)

        self.setAlignment(Qt.AlignCenter)
        tw = self.fontMetrics().width(" NORMAL ")
        fw = self.style().pixelMetric(self.style().PM_DefaultFrameWidth)
        self.setFixedWidth(tw + (2 * fw) + 4)

    @Slot(int)
    def change_state(self, state):
        """Display the state of vim."""
        try:
            self.setStyleSheet("QLabel { color: white }")
            if state == VimState.VISUAL:
                self.setText("VISUAL")
                self.setStyleSheet("QLabel { background-color: #ff8000 }")
            elif state == VimState.NORMAL:
                self.setText("NORMAL")
                self.setStyleSheet("QLabel { background-color: #29a329 }")
            elif state == VimState.VLINE:
                self.setText("V-LINE")
                self.setStyleSheet("QLabel { background-color: #ff8000 }")
            elif state == VimState.INSERT:
                self.setText("INSERT")
                self.setStyleSheet("QLabel { background-color: #3366ff }")
        except RuntimeError:
            # Ignore updates to deleted widgets
            pass


class VimLineEdit(QLineEdit):
    """Vim Command input."""

    def __init__(self, vim_widget, vim_status: VimStatus, vim_shortcut: VimShortcut):
        super().__init__(vim_widget)
        self.vim_widget = vim_widget
        self.vim_status = vim_status
        self.vim_shortcut = vim_shortcut

        # Set size
        tw = self.fontMetrics().width(" :%s/international/internationl/g ")
        fw = self.style().pixelMetric(self.style().PM_DefaultFrameWidth)
        self.setFixedWidth(tw + (2 * fw) + 4)

        # Todo: Move setting shortcut to config file.
        vim_shortcut.signal_cmd.connect(self.setText)
        self.dispatcher = {
            Qt.Key_A: vim_shortcut.add_num,
            Qt.Key_X: vim_shortcut.subtract_num,
            Qt.Key_D: vim_shortcut.pg_half_down,
            Qt.Key_F: vim_shortcut.pg_down,
            Qt.Key_U: vim_shortcut.pg_half_up,
            Qt.Key_B: vim_shortcut.pg_up,
            Qt.Key_R: vim_shortcut.redo,
            Qt.Key_P: vim_shortcut.open_path_finder,
            Qt.Key_C: vim_shortcut.clear_tip_search,
        }
        self.setAttribute(Qt.WA_InputMethodEnabled, False)

    def to_normal(self):
        """Convert the state of vim to normal mode."""
        self.clear()
        self.vim_widget.vim_status.to_normal()
        self.vim_widget.vim_status.cursor.draw_vim_cursor()

    def keyPressEvent(self, e):
        """Override Qt method."""
        self.vim_status.manager_macro.add_vim_keyevent(e)

        key = e.key()
        pressed_ctrl = e.modifiers() == Qt.ControlModifier
        if key == Qt.Key_Escape:
            self.esc_pressed()
        elif KEYCODE2STR.get(key, None):
            self.setText(self.text() + KEYCODE2STR[key])
        elif pressed_ctrl and key in self.dispatcher.keys():
            self.dispatcher[key]()
        else:
            super().keyPressEvent(e)

    def esc_pressed(self):
        """Clear state."""
        self.vim_status.input_cmd.clear()
        self.vim_status.remove_marker_of_easymotion()
        if self.vim_status.sub_mode:
            self.vim_status.sub_mode = None
            self.clear()
        else:
            self.to_normal()

    def focusInEvent(self, event):
        """Override Qt method."""
        self.vim_status.disconnect_from_editor()
        super().focusInEvent(event)
        if self.vim_status.cursor.get_editor():
            self.to_normal()
        self.vim_status.set_message("")

    def focusOutEvent(self, event):
        """Override Qt method."""
        super().focusOutEvent(event)
        self.clear()
        try:
            self.vim_widget.vim_status.to_insert()
        except RuntimeError:
            # The status may already be deleted when closing widgets
            pass


class VimMsgLabel(QLabel):
    """Display message of vim."""

    def __init__(self, txt, parent):
        super().__init__(txt, parent)
        self.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        tw = self.fontMetrics().width(" recording @q... 0000 fewers lines ")
        fw = self.style().pixelMetric(self.style().PM_DefaultFrameWidth)
        self.setFixedWidth(tw + (2 * fw) + 4)


class WorkerMacro(QThread):
    """Send keyinfo from registers to main thread."""

    sig_focus_vim = Signal()
    sig_send_key_info = Signal(object)

    def __init__(self, parent):
        super().__init__(parent)
        self.key_info_list = []
        self.num_iteration = 0

    def set_key_infos(self, key_infos, num):
        """Set key infos of registers."""
        self.key_info_list = key_infos.copy()
        self.num_iteration = num

    @coverage_resolve_trace
    def run(self):
        """Send key info to main thread."""
        is_focus_vim = True
        for _ in range(self.num_iteration):
            for key_info in self.key_info_list:
                if key_info.identifier == 1:
                    is_focus_vim = False
                elif key_info.identifier == 0 and is_focus_vim is False:
                    is_focus_vim = True
                    self.sig_focus_vim.emit()
                self.sig_send_key_info.emit(key_info)

        if is_focus_vim is False:
            self.sig_focus_vim.emit()


class VimWidget(QWidget):
    """Vim widget."""

    def __init__(self, editor_widget, main):
        super().__init__(main)
        self.editor_widget = editor_widget
        self.main = main
        self.status_label = VimStateLabel(main)
        self.msg_label = VimMsgLabel("", main)

        self.vim_status = VimStatus(editor_widget, main, self.msg_label)
        self.vim_status.change_label.connect(self.status_label.change_state)

        self.vim_shortcut = VimShortcut(self.main, self.vim_status)

        self.commandline = VimLineEdit(self, self.vim_status, self.vim_shortcut)
        self.commandline.textChanged.connect(self.on_text_changed)

        self.vim_status.cmd_line = self.commandline
        self.vim_shortcut.cmd_line = self.commandline

        self.executor_normal_cmd = ExecutorNormalCmd(self.vim_status)
        self.executor_visual_cmd = ExecutorVisualCmd(self.vim_status)
        self.executor_vline_cmd = ExecutorVlineCmd(self.vim_status)
        self.executors = {
            VimState.NORMAL: self.executor_normal_cmd,
            VimState.VISUAL: self.executor_visual_cmd,
            VimState.VLINE: self.executor_vline_cmd,
        }

        # leader key
        self.executor_leader_key = ExecutorLeaderKey(self.vim_status)
        self.leader_key = " "
        self.set_leader_key()

        # macro
        self.worker_macro = WorkerMacro(main)
        self.worker_macro.sig_send_key_info.connect(self.send_key_event)
        self.worker_macro.sig_focus_vim.connect(self.commandline.setFocus)

    @Slot(object)
    def send_key_event(self, key_info):
        event = key_info.to_event()
        if key_info.identifier == 0:
            self.commandline.keyPressEvent(event)
        else:
            editor = self.vim_status.get_editor()
            editor.keyPressEvent(event)

    def set_leader_key(self):
        """Set leader key from CONF."""
        leader_key = CONF.get(CONF_SECTION, "leader_key")
        leader_key2 = KEYCODE2STR.get(QKeySequence.fromString(leader_key)[0], None)
        if leader_key2:
            leader_key = leader_key2
        self.leader_key = leader_key
        self.executor_leader_key.set_easymotion_key(self.leader_key)

    def on_text_changed(self, txt):
        """Send input command to executor."""
        if not txt:
            return

        executor = self.executors[self.vim_status.vim_state]
        if self.vim_status.sub_mode:
            executor = self.vim_status.sub_mode

        # workaround for easymotion (press leader, leader)
        if txt == self.leader_key and executor != self.executor_leader_key:
            sub_mode = self.vim_status.sub_mode
            if sub_mode and sub_mode.allow_leaderkey is False:
                pass
            else:
                self.executor_leader_key.prev_executor = executor
                self.vim_status.sub_mode = self.executor_leader_key
                self.commandline.clear()
                return

        if executor(txt):
            self.commandline.clear()

        if self.vim_status.manager_macro.reg_name_for_execute:
            mm = self.vim_status.manager_macro
            ch = mm.reg_name_for_execute
            self.worker_macro.set_key_infos(mm.registers[ch], mm.num_execute)
            self.worker_macro.start()
            mm.set_info_for_execute("", 0)



# -*- coding: utf-8 -*-
"""Status of Vim."""
# Standard library imports
from collections import defaultdict
from typing import List

# Third party imports
from qtpy.QtCore import (
    QEvent,
    QObject,
    Qt,
    QRegularExpression,
    QTimer,
    Signal,
    Slot,
)
from qtpy.QtGui import (
    QBrush,
    QColor,
    QKeyEvent,
    QRegularExpressionValidator,
    QTextCursor,
    QValidator,
)
from qtpy.QtWidgets import QApplication, QLabel, QTextEdit
from spyder.config.manager import CONF
from spyder.plugins.editor.api.decoration import DRAW_ORDERS

# Local imports
from spyder_okvim.spyder.config import CONF_SECTION
from spyder_okvim.utils.easymotion import ManageMarkerEasymotion, PainterEasyMotion
from spyder_okvim.utils.helper_motion import MotionInfo, MotionType


class VimState:
    """Vim state constants."""

    VISUAL = 0
    VLINE = 1
    NORMAL = 3
    INSERT = 4
    SUBMODE = 5


class FindInfo:
    """f,F,t,T command Info."""

    def __init__(self):
        self.cmd_name: str = ""
        self.ch: str = ""

    def set(self, cmd_name: str, ch: str):
        """Set infomation of find commmand."""
        self.cmd_name = cmd_name
        self.ch = ch


class InputCmdInfo:
    """Info for input command."""

    def __init__(self, num_str: str, cmd: str):
        self.num_str: str = num_str
        self.cmd: str = cmd

    def clear(self):
        """Clear."""
        self.num_str = ""
        self.cmd = ""

    def set(self, input_cmd_info):
        """Set the information of input command."""
        self.num_str = input_cmd_info.num_str
        self.cmd = input_cmd_info.cmd


class DotCmdInfo:
    """Dot command info."""

    def __init__(self):
        self.vim_state = None
        self.sel_block_no_start = None
        self.sel_block_no_end = None
        self.sel_col_start = None
        self.sel_col_end = None
        self.num_str: str = ""
        self.cmd: str = ""
        self.cmd_list_insertmode = None
        self.register_name = None
        self.editor_connected = None
        self.key_list_from_editor = []
        self.key_list_to_cmd_line = []

    def clear_key_list(self):
        """Clear key list from editor."""
        self.key_list_from_editor.clear()
        self.key_list_to_cmd_line.clear()

    def cmd2string(self, num, num_str):
        """Convert cmd info to string of vim command."""
        # TODO : apply register name.
        cmd_str = ""

        if self.vim_state == VimState.VLINE:
            cmd_str += "V"
            n_lines = self.sel_block_no_end - self.sel_block_no_start
            if n_lines > 0:
                cmd_str += f"{n_lines}j"
            cmd_str += self.cmd
            return cmd_str

        if self.vim_state == VimState.VISUAL:
            cmd_str += "v"
            n_lines = self.sel_block_no_end - self.sel_block_no_start
            if n_lines > 0:
                cmd_str += f"{n_lines}j0"

                if self.sel_col_end > 1:
                    cmd_str += f"{self.sel_col_end - 1}l"
            else:
                n = self.sel_col_end - self.sel_col_start - 1
                if n > 0:
                    cmd_str += f"{n}l"
            cmd_str += self.cmd
            return cmd_str

        if num_str:
            return str(num) + self.cmd
        else:
            return self.num_str + self.cmd


class KeyInfo:
    """Save the info of QKeyEvent."""

    def __init__(self, key_code: int, text: str, modifiers: int, identifier: int = 1):
        self.key_code = key_code
        self.text = text
        self.modifiers = modifiers
        self.identifier = identifier  # 0: from vim, 1: from editor

    def to_event(self):
        """Convert key info to qkeyevent."""
        event = QKeyEvent(QEvent.KeyPress, self.key_code, self.modifiers, self.text)
        event.ignore()
        return event


class RegisterInfo:
    """Info of register."""

    def __init__(self):
        self.name = ""
        self.content = ""
        self.type = VimState.NORMAL


class SearchInfo:
    """Result of search."""

    def __init__(self, vim_cursor):
        self.color_fg = QBrush(QColor("#A9B7C6"))
        self.color_bg = QBrush(QColor("#30652F"))
        self.txt_searched = ""
        self.selection_list = []
        self.vim_cursor = vim_cursor
        self.ignorecase = True

        self.set_color()

    def set_color(self):
        """Set the color of search results."""
        self.color_fg = QBrush(QColor(CONF.get(CONF_SECTION, "search_fg_color")))
        self.color_bg = QBrush(QColor(CONF.get(CONF_SECTION, "search_bg_color")))

        for sel in self.selection_list:
            sel.format.setForeground(self.color_fg)
            sel.format.setBackground(self.color_bg)

    def get_sel_start_list(self):
        """Get the start position of selection."""
        cursor = self.vim_cursor.get_cursor()
        validator = QRegularExpressionValidator(
            QRegularExpression(self.txt_searched), None
        )

        tmp = []
        for sel in self.selection_list:
            cursor.setPosition(sel.cursor.selectionStart())
            cursor.setPosition(sel.cursor.selectionEnd(), QTextCursor.KeepAnchor)
            txt_sel = cursor.selectedText()

            if (validator.validate(txt_sel, 0)[0] == QValidator.Acceptable) or (
                self.ignorecase is True and txt_sel.lower() == self.txt_searched.lower()
            ):
                tmp.append(sel)
        self.selection_list = tmp

        self.vim_cursor.set_extra_selections(
            "vim_search", [i for i in self.selection_list]
        )

        return [i.cursor.selectionStart() for i in self.selection_list]


class ManagerMacro:
    """Manage macros."""

    def __init__(self):
        self.registers = defaultdict(list)
        self.is_recording = False
        self.reg_name_for_record = ""
        self.editor_connected = None
        self.reg_name_for_execute = ""
        self.num_execute = 0

    def set_info_for_execute(self, ch, num):
        """Set info for executing contents of register."""
        self.reg_name_for_execute = ch
        self.num_execute = num

    def start_record(self, ch):
        """Start recording."""
        self.reg_name_for_record = ch
        self.is_recording = True

        if ch.lower() == ch:
            self.registers[ch] = list()

    def stop_record(self):
        """Stop recording."""
        # remove the last key if the last key is q
        self.remove_last_key("q")
        self.reg_name_for_record = ""
        self.is_recording = False

    def remove_last_key(self, ch):
        """Remove last key."""
        key_list = self.registers[self.reg_name_for_record]
        if len(key_list) > 0:
            last_key = key_list[-1]
            if last_key.text == ch:
                self.registers[self.reg_name_for_record] = key_list[:-1]

    def add_vim_keyevent(self, event):
        """Add keyevent from vim."""
        if self.is_recording:
            self.registers[self.reg_name_for_record].append(
                KeyInfo(event.key(), event.text(), event.modifiers(), 0)
            )

    def add_editor_keyevent(self, event):
        """Add keyevent from editor."""
        self.registers[self.reg_name_for_record].append(
            KeyInfo(event.key(), event.text(), event.modifiers(), 1)
        )


class LabelOnTxt(QLabel):
    """Label on txt."""

    def __init__(self, parent=None):
        """."""
        super().__init__(parent)
        self.setWindowFlags(Qt.FramelessWindowHint | Qt.WindowStaysOnTopHint)

    def set_style(self, font_family: str, font_size_pt: int):
        """Set stylesheet."""
        self.setStyleSheet(
            f"""QLabel {{ 
            background-color : #222b35;
            color : #cd4340;
            border: 1px solid #cd4340;
            padding : 0px;
            text-indent : 1px; 
            margin : 0px;
            font-family: {font_family};
            font-size: {font_size_pt}pt;
            }}"""
        )


class VimCursor:
    """Manange the cursor in vim."""

    def __init__(self, editor_widget):
        self.editor_widget = editor_widget

        self.vim_cursor = QTextEdit.ExtraSelection()
        self.vim_cursor.format.setForeground(QBrush(QColor("#000000")))
        self.vim_cursor.format.setBackground(QBrush(QColor("#BBBBBB")))

        self.selection = QTextEdit.ExtraSelection()
        self.selection.format.setForeground(QBrush(QColor("#A9B7C6")))
        self.selection.format.setBackground(QBrush(QColor("#214283")))

        self.yank_fg_color = QBrush(QColor("#B9C7D6"))
        self.yank_bg_color = QBrush(QColor("#7D7920"))
        self.hl_yank_dur = 400  # duration of highlight after yank
        self.hl_yank = True

        self.set_config_from_conf()

        # Order of Selections
        DRAW_ORDERS["vim_search"] = 6
        DRAW_ORDERS["hl_yank"] = 7
        DRAW_ORDERS["vim_selection"] = 8
        DRAW_ORDERS["vim_cursor"] = 9

    def set_config_from_conf(self):
        """Set config from conf."""
        self.vim_cursor.format.setForeground(
            QBrush(QColor(CONF.get(CONF_SECTION, "cursor_fg_color")))
        )
        self.vim_cursor.format.setBackground(
            QBrush(QColor(CONF.get(CONF_SECTION, "cursor_bg_color")))
        )

        self.selection.format.setForeground(
            QBrush(QColor(CONF.get(CONF_SECTION, "select_fg_color")))
        )
        self.selection.format.setBackground(
            QBrush(QColor(CONF.get(CONF_SECTION, "select_bg_color")))
        )

        self.yank_fg_color = QBrush(QColor(CONF.get(CONF_SECTION, "yank_fg_color")))
        self.yank_bg_color = QBrush(QColor(CONF.get(CONF_SECTION, "yank_bg_color")))
        self.hl_yank_dur = CONF.get(CONF_SECTION, "highlight_yank_duration")
        self.hl_yank = CONF.get(CONF_SECTION, "highlight_yank")

    def get_editor(self):
        """Get the editor focused."""
        editorstack = self.editor_widget.get_current_editorstack()
        return editorstack.get_current_editor()

    def get_editorstack(self):
        """Get the editorstack."""
        return self.editor_widget.get_current_editorstack()

    def get_cursor(self):
        """Get the cursor."""
        return self.get_editor().textCursor()

    def set_cursor(self, cursor):
        """Set the cursor to focused editor."""
        return self.get_editor().setTextCursor(cursor)

    def get_end_position(self):
        """Get the end position of document."""
        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.End)
        return cursor.position()

    def draw_vim_cursor(self):
        """Draw vim cursor."""
        vim_cursor = self.vim_cursor
        editor = self.get_editor()
        vim_cursor.cursor = editor.textCursor()
        vim_cursor.cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_cursor", [vim_cursor])
        # editor.update()

    def create_selection(self, start, end):
        """Create selection."""
        end_document = self.get_end_position()
        if end > end_document:
            end = end_document

        sel = self.selection
        sel.cursor = QTextCursor(self.get_editor().document())
        sel.cursor.setPosition(start)
        sel.cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_selection", [sel])
        self.draw_vim_cursor()

    def set_cursor_pos_in_visual(self, pos_new):
        """Set cursor position and redraw selection in visual mode.

        Parameters
        ----------
        pos_new : int
            position of cursor

        """
        if pos_new is None:
            return
        editor = self.get_editor()
        pos_cur = editor.textCursor().position()

        sel = editor.get_extra_selections("vim_selection")[0]
        start_old = sel.cursor.selectionStart()
        end_old = sel.cursor.selectionEnd()

        start = start_old
        end = end_old

        if abs(start_old - pos_cur) <= abs(end_old - 1 - pos_cur):
            if start_old >= pos_new:
                start = pos_new
            elif end_old > pos_new:
                start = pos_new
            else:
                start = end - 1
                end = pos_new + 1
        else:
            if end_old - 1 <= pos_new:
                end = pos_new + 1
            elif start_old < pos_new:
                end = pos_new + 1
            else:
                end = start_old + 1
                start = pos_new

        end_document = self.get_end_position()
        if end > end_document:
            end = end_document

        sel.cursor.setPosition(start)
        sel.cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_selection", [sel])
        self.set_cursor_pos(pos_new)

    def get_block(self, pos):
        """Get block number of cursor position.

        Returns
        -------
        QTextBlock
            The QTextBlock class provides a container for text fragments.
        int
            Block number.
        """
        cursor = self.get_cursor()
        cursor.setPosition(pos)
        return cursor.block(), cursor.blockNumber()

    def get_block_no_start_in_selection(self):
        """Get start block number of selection."""
        editor = self.get_editor()

        sel = editor.get_extra_selections("vim_selection")
        if not sel:
            return

        sel = sel[0]

        sel_start = sel.cursor.selectionStart()

        _, block_no_start = self.get_block(sel_start)

        return block_no_start

    def get_block_no_end_in_selection(self):
        """Get the last block number of selection."""
        editor = self.get_editor()

        sel = editor.get_extra_selections("vim_selection")
        if not sel:
            return

        sel = sel[0]

        sel_end = sel.cursor.selectionEnd()

        _, block_no_end = self.get_block(sel_end)

        return block_no_end

    def get_pos_start_in_selection(self):
        """Get start position of selection."""
        editor = self.get_editor()

        sel = editor.get_extra_selections("vim_selection")
        if not sel:
            return

        sel = sel[0]

        sel_start = sel.cursor.selectionStart()

        return sel_start

    def get_pos_end_in_selection(self):
        """Get end position of selection."""
        editor = self.get_editor()

        sel = editor.get_extra_selections("vim_selection")
        if not sel:
            return

        sel = sel[0]

        sel_end = sel.cursor.selectionEnd()

        return sel_end

    def set_selection_to_editor_using_vim_selection(self):
        """Set the selection to editor by referring the selection of vim."""
        editor = self.get_editor()
        old_cursor = editor.textCursor()
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()
        if sel_start:
            new_cursor = editor.textCursor()
            new_cursor.setPosition(sel_start, QTextCursor.MoveAnchor)
            new_cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
            editor.setTextCursor(new_cursor)
            return old_cursor

    def set_cursor_pos_in_vline(self, pos_new):
        """Set cursor position and redraw selection in vline mode.

        Parameters
        ----------
        pos_new : int
            position of cursor

        """
        if pos_new is None:
            return
        editor = self.get_editor()
        block_no_cur = editor.textCursor().blockNumber()

        sel = editor.get_extra_selections("vim_selection")[0]
        sel_start = sel.cursor.selectionStart()
        sel_end = sel.cursor.selectionEnd()

        block_start, block_no_start = self.get_block(sel_start)
        block_end, block_no_end = self.get_block(sel_end)
        block_new, block_no_new = self.get_block(pos_new)

        if block_no_start == block_no_cur:
            if block_no_new <= block_no_end:
                block_start = block_new
            else:
                block_start = block_end
                block_end = block_new
        else:
            if block_no_new >= block_no_start:
                block_end = block_new
            else:
                block_end = block_start
                block_start = block_new

        start = block_start.position()
        end = block_end.position() + block_end.length() - 1
        sel.cursor.setPosition(start)
        sel.cursor.setPosition(end, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_selection", [sel])
        self.set_cursor_pos(pos_new)

    def set_block_selection_in_visual(self, motion_info: MotionInfo):
        """Set block selection in visual."""
        editor = self.get_editor()
        sel = editor.get_extra_selections("vim_selection")[0]
        sel.cursor.setPosition(motion_info.sel_start)
        sel.cursor.setPosition(motion_info.sel_end, QTextCursor.KeepAnchor)
        self.set_extra_selections("vim_selection", [sel])
        self.set_cursor_pos(motion_info.sel_end - 1)

    def set_cursor_pos(self, pos):
        """Set position cursor of editor."""
        if pos is None:
            return
        editor = self.get_editor()
        cursor = editor.textCursor()
        cursor.setPosition(pos)
        self.set_cursor(cursor)
        self.draw_vim_cursor()

    def set_cursor_pos_without_end(self, pos):
        """Set position cursor of editor.

        If the new position is the end of the block, cusor moves to left.
        """
        if pos is None:
            return
        editor = self.get_editor()
        cursor = editor.textCursor()
        cursor.setPosition(pos)
        if cursor.atBlockEnd() and not cursor.atBlockStart():
            cursor.movePosition(QTextCursor.Left)
        self.set_cursor(cursor)
        self.draw_vim_cursor()

    def apply_motion_info_in_normal(self, motion_info: MotionInfo):
        """Apply motion info in normal mode."""
        self.set_cursor_pos(motion_info.cursor_pos)

    def apply_motion_info_in_yank(self, motion_info: MotionInfo):
        """Apply motion info after yank in normal mode."""
        cursor_pos = self.get_cursor().position()
        if motion_info.motion_type == MotionType.BlockWise:
            self.set_cursor_pos(motion_info.sel_start)
        else:
            if motion_info.cursor_pos is None:
                return
            pos = min([cursor_pos, motion_info.cursor_pos])
            self.set_cursor_pos(pos)

    def apply_motion_info_in_visual(self, motion_info: MotionInfo):
        """Apply motion info in visual mode."""
        self.set_cursor_pos_in_visual(motion_info.cursor_pos)

    def apply_motion_info_in_vline(self, motion_info: MotionInfo):
        """Apply motion info in visual mode."""
        self.set_cursor_pos_in_vline(motion_info.cursor_pos)

    def set_extra_selections(self, key, sels):
        """Set the selection info to editor."""
        editor = self.get_editor()
        editor.set_extra_selections(key, sels)

    def highlight_yank(self, pos_start, pos_end):
        """Highlight after yank."""
        if self.hl_yank is False:
            return

        cursor = self.get_cursor()
        editor = self.get_editor()

        sel = QTextEdit.ExtraSelection()
        sel.format.setForeground(self.yank_fg_color)
        sel.format.setBackground(self.yank_bg_color)
        sel.cursor = cursor

        sel.cursor.setPosition(pos_start)
        sel.cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        self.set_extra_selections("hl_yank", [sel])

        QTimer.singleShot(
            self.hl_yank_dur, lambda: editor.clear_extra_selections("hl_yank")
        )


class VimStatus(QObject):
    """Status of vim."""

    change_label = Signal(int)

    def __init__(self, editor_widget, main, msg_label):
        super().__init__()
        self.is_visual_mode = False
        self.vim_state = VimState.NORMAL
        self.editor_widget = editor_widget
        self.cursor: VimCursor = VimCursor(editor_widget)
        self.main = main

        # method mapping
        self.set_cursor = self.cursor.set_cursor
        self.get_cursor = self.cursor.get_cursor
        self.get_editor = self.cursor.get_editor
        self.get_editorstack = self.cursor.get_editorstack
        self.get_block_no_start_in_selection = (
            self.cursor.get_block_no_start_in_selection
        )
        self.get_block_no_end_in_selection = self.cursor.get_block_no_end_in_selection
        self.get_pos_start_in_selection = self.cursor.get_pos_start_in_selection
        self.get_pos_end_in_selection = self.cursor.get_pos_end_in_selection

        self.cmd_line = None

        self.sub_mode = None

        # command
        self.find_info = FindInfo()
        self.input_cmd = InputCmdInfo("", "")
        self.input_cmd_prev = InputCmdInfo("", "")
        self.dot_cmd = DotCmdInfo()
        self.running_dot_cmd = False

        # register
        self.register_dict = defaultdict(RegisterInfo)

        # config
        self.indent = "    "

        # search
        self.search = SearchInfo(self.cursor)

        # message
        self.msg_label = msg_label
        self.msg_prefix = ""

        # Macro
        self.manager_macro = ManagerMacro()

        # easymotion
        self.painter_easymotion = PainterEasyMotion()
        self.manager_marker_easymotion = ManageMarkerEasymotion()
        self.editor_connected_easymotion = None

        # Sneak
        self.n_annotate_max = 50
        self.labels_for_annotate = [
            LabelOnTxt(self.main) for _ in range(self.n_annotate_max)
        ]
        self.hide_annotate_on_txt()

    def clear_state(self):
        """Clear."""
        self.is_visual_mode = False
        self.vim_state = VimState.NORMAL
        editor = self.get_editor()
        if editor:
            editor.clear_extra_selections("vim_selection")
            editor.clear_extra_selections("vim_cursor")

    def is_normal(self):
        """Check that vim state is normal mode."""
        return self.vim_state == VimState.NORMAL

    def to_normal(self):
        """Change vim state to normal mode."""
        self.clear_state()
        cursor = self.get_cursor()
        if cursor.atBlockEnd() and not cursor.atBlockStart():
            pos = cursor.position()
            self.cursor.set_cursor_pos(pos - 1)
        self.change_label.emit(VimState.NORMAL)

    def reset_for_test(self):
        """Reset status for test."""
        self.clear_state()

        self.find_info = FindInfo()
        self.input_cmd = InputCmdInfo("", "")
        self.input_cmd_prev = InputCmdInfo("", "")
        self.dot_cmd.clear_key_list()
        self.running_dot_cmd = False

        self.get_editor().clear_extra_selections("hl_yank")

        # register
        self.register_dict = defaultdict(RegisterInfo)

        # config
        self.indent = "    "

        # search
        self.search = SearchInfo(self.cursor)

        # Macro
        self.manager_macro = ManagerMacro()

        self.msg_label.setText("")

        self.to_normal()

    def to_insert(self):
        """Change vim state to insert mode."""
        self.is_visual_mode = False
        self.remove_marker_of_easymotion()
        self.get_editor().clear_extra_selections("vim_selection")
        self.get_editor().clear_extra_selections("vim_cursor")
        self.change_label.emit(VimState.INSERT)

    def to_visual_char(self):
        """Change vim state to visual mode."""
        self.clear_state()
        self.is_visual_mode = True
        self.vim_state = VimState.VISUAL

        pos = self.get_cursor().position()
        self.cursor.create_selection(pos, pos + 1)

        self.change_label.emit(VimState.VISUAL)

    def to_visual_line(self):
        """Change vim state to vline mode."""
        self.clear_state()
        self.is_visual_mode = True
        self.vim_state = VimState.VLINE

        cursor = self.get_cursor()
        cursor.movePosition(QTextCursor.StartOfBlock)

        pos_start = cursor.position()

        cursor.movePosition(QTextCursor.EndOfBlock)
        pos_end = cursor.position()

        self.cursor.create_selection(pos_start, pos_end)

        self.change_label.emit(VimState.VLINE)

    def get_number_of_visible_lines(self):
        """Get the number of visible lines in editor."""
        editor = self.get_editor()
        num_lines = editor.viewport().height() // editor.fontMetrics().height()
        return num_lines

    def set_focus_to_vim(self):
        """Set focus to vim command line."""
        if self.cmd_line:
            self.cmd_line.setFocus()

    def set_focus_to_vim_after_delay(self, delay=300):
        """Set focus to vim command line after delay.

        Spyder is focused after receiving the result of pyls.
        So set_focus_to_vim method is useless.
        """

        def focus():
            self.set_focus_to_vim()
            self.cursor.draw_vim_cursor()

        if self.cmd_line:
            QTimer.singleShot(delay, focus)

    @Slot(QKeyEvent)
    def rcv_key_from_editor(self, event):
        """Add key event from editor to list."""
        self.dot_cmd.key_list_from_editor.append(
            KeyInfo(event.key(), event.text(), event.modifiers(), 1)
        )

    def disconnect_from_editor(self):
        """Disconnect from the editor."""
        # disconnect previous connection.
        editor = self.dot_cmd.editor_connected
        if editor:
            editor.sig_key_pressed.disconnect(self.rcv_key_from_editor)
        self.dot_cmd.editor_connected = None

    def update_dot_cmd(
        self, connect_editor, register_name=None, key_list_to_cmd_line=None
    ):
        """Update input command info to dot command info."""
        if self.running_dot_cmd is True:
            return
        dot_cmd = self.dot_cmd
        dot_cmd.vim_state = self.vim_state
        dot_cmd.num_str = self.input_cmd.num_str
        dot_cmd.cmd = self.input_cmd.cmd
        dot_cmd.register_name = register_name
        dot_cmd.clear_key_list()
        if key_list_to_cmd_line:
            dot_cmd.key_list_to_cmd_line = key_list_to_cmd_line
        self.disconnect_from_editor()

        # For receiving key event from codeeditor
        if connect_editor:
            editor = self.get_editor()
            self.dot_cmd.editor_connected = editor
            editor.sig_key_pressed.connect(self.rcv_key_from_editor)

        # Get selection region.
        if self.is_normal():
            return

        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()

        block_start, block_no_start = self.cursor.get_block(sel_start)
        block_end, block_no_end = self.cursor.get_block(sel_end)

        dot_cmd.sel_block_no_start = block_no_start
        dot_cmd.sel_block_no_end = block_no_end
        dot_cmd.sel_col_start = sel_start - block_start.position()
        dot_cmd.sel_col_end = sel_end - block_end.position()

    def get_register_name(self):
        """Get register name from previous command."""
        txt = self.input_cmd_prev.cmd

        if len(txt) == 2 and txt[0] == '"':
            return txt[1]
        else:
            return '"'  # default register name

    def set_register(self, name, content, register_type):
        """Set content into register_dict."""
        if name == "+":
            QApplication.clipboard().setText(content)
        else:
            info = self.register_dict[name]
            info.name = name
            info.content = content
            info.type = register_type
            self.register_dict[name] = info

    def get_register(self):
        """Get content from register_dict."""
        name = self.get_register_name()
        if name == "+":
            info = RegisterInfo()
            info.name = "+"
            info.content = QApplication.clipboard().text()
            info.type = VimState.NORMAL
            return info
        else:
            return self.register_dict[name]

    def set_message(self, msg, duration_ms=-1):
        """Display the massage."""
        self.msg_label.setText(f"{self.msg_prefix}{msg}")

    def start_recording_macro(self, reg_name):
        """Start recording macro."""
        self.msg_prefix = f"recording @{reg_name}... "
        self.manager_macro.start_record(reg_name)

        editor = self.get_editor()
        self.manager_macro.editor_connected = editor
        editor.sig_key_pressed.connect(self.add_key_from_editor_to_macro_manager)

    def is_recording_macro(self):
        """Return is_recording."""
        return self.manager_macro.is_recording

    def stop_recoding_macro(self):
        """Stop recording macro."""
        self.msg_prefix = ""
        self.manager_macro.stop_record()
        self.disconnect_macro_manager_from_editor()

    @Slot(QKeyEvent)
    def add_key_from_editor_to_macro_manager(self, event):
        """Add key event from editor to list to macro_manager."""
        self.manager_macro.add_editor_keyevent(event)

    def disconnect_macro_manager_from_editor(self):
        """Disconnect macro manager from the editor."""
        # disconnect previous connection.
        editor = self.manager_macro.editor_connected
        if editor:
            editor.sig_key_pressed.disconnect(self.add_key_from_editor_to_macro_manager)
        self.manager_macro.editor_connected = None

    def set_marker_for_easymotion(self, positions: List[int], motion_type):
        """Set marker for easymotion."""
        if not positions:
            return
        self.remove_marker_of_easymotion()
        editor = self.get_editor()
        self.manager_marker_easymotion.set_positions(positions, motion_type)
        self.painter_easymotion.editor = editor
        self.painter_easymotion.positions = positions
        self.painter_easymotion.names = self.manager_marker_easymotion.name_list
        self.editor_connected_easymotion = editor
        editor.viewport().installEventFilter(self.painter_easymotion)
        editor.viewport().update()

    def update_marker_for_easymotion(self):
        """Update marker of easymotion."""
        self.remove_marker_of_easymotion()
        editor = self.get_editor()
        self.painter_easymotion.editor = editor
        self.painter_easymotion.positions = self.manager_marker_easymotion.position_list
        self.painter_easymotion.names = self.manager_marker_easymotion.name_list
        self.editor_connected_easymotion = editor
        editor.viewport().installEventFilter(self.painter_easymotion)
        editor.viewport().update()

    def remove_marker_of_easymotion(self):
        """Remove marker of easymotion."""
        editor = self.editor_connected_easymotion
        self.editor_connected_easymotion = None
        if editor:
            editor.viewport().removeEventFilter(self.painter_easymotion)
            editor.viewport().update()

    @Slot()
    def hide_annotate_on_txt(self):
        """Hide Labels for annotate on txt."""
        for idx in range(self.n_annotate_max):
            self.labels_for_annotate[idx].hide()

    def annotate_on_txt(self, info: dict[int, str], timeout: int = -1):
        """Annotate on txt."""
        editor = self.get_editor()
        tc = editor.textCursor()
        font = editor.viewport().font()
        fm = editor.fontMetrics()
        ch_width = fm.width("H")
        ch_height = fm.height()

        for idx, (pos, data) in enumerate(info.items()):
            label = self.labels_for_annotate[idx]
            label.set_style(font.family(), font.pointSize())
            label.setFixedSize(ch_width * len(data), ch_height)
            label.setText(data)

            tc.setPosition(pos)
            rect_in_local = editor.cursorRect(tc)
            pos_in_parent = editor.viewport().mapTo(
                label.parent(), rect_in_local.topLeft()
            )
            print(pos_in_parent)
            label.move(pos_in_parent)
            label.raise_()
            label.show()

            if idx > self.n_annotate_max - 1:
                break

        QTimer.singleShot(timeout, self.hide_annotate_on_txt)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#
"""Tests for the plugin."""
# Third party imports
import pytest
from qtpy.QtCore import QEvent, Qt
from qtpy.QtGui import QKeyEvent, QFocusEvent

# Local imports
from spyder_okvim.spyder.confpage import OkvimConfigPage


def test_conf_page(vim_bot):
    """Test conf_page.

    Call the methods that is difficult to make test case.
    """
    main, _, _, vim, _ = vim_bot
    vim.get_icon()
    vim.switch_to_plugin()
    conf_page = OkvimConfigPage(vim, main)
    conf_page.setup_page()

    old_leader_key = conf_page.leaderkey_viewer.textbox.text()
    new_event = QKeyEvent(QEvent.KeyPress, Qt.Key_Control, Qt.NoModifier)
    conf_page.leaderkey_edit.keyPressEvent(new_event)

    assert conf_page.leaderkey_viewer.textbox.text() == old_leader_key

    new_event = QKeyEvent(QEvent.KeyPress, Qt.Key_unknown, Qt.NoModifier)
    conf_page.leaderkey_edit.keyPressEvent(new_event)

    assert conf_page.leaderkey_viewer.textbox.text() == old_leader_key

    new_event = QKeyEvent(QEvent.KeyPress, Qt.Key_Meta, Qt.NoModifier)
    conf_page.leaderkey_edit.keyPressEvent(new_event)

    assert conf_page.leaderkey_viewer.textbox.text() == old_leader_key

    new_event = QKeyEvent(QEvent.KeyPress, Qt.Key_Enter, Qt.NoModifier)
    conf_page.leaderkey_edit.keyPressEvent(new_event)

    assert conf_page.leaderkey_viewer.textbox.text() == "Enter"


def test_apply_config(vim_bot):
    """Run apply_plugin_settings method."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("foo Foo foo Foo")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    # test refresh the color of search result.
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "/foo")
    qtbot.keyPress(cmd_line, Qt.Key_Return)

    vim.apply_plugin_settings("")


def test_ctrl_u_b(vim_bot):
    """Test ^u ^b."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "j")

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_U, Qt.ControlModifier)
    vim.vim_cmd.commandline.keyPressEvent(event)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == 0

    qtbot.keyClicks(cmd_line, "j")

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_B, Qt.ControlModifier)
    vim.vim_cmd.commandline.keyPressEvent(event)

    assert cmd_line.text() == ""
    assert editor.textCursor().position() == 0


def test_ctrl_d_f(vim_bot):
    """Test ^d ^f."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_D, Qt.ControlModifier)
    vim.vim_cmd.commandline.keyPressEvent(event)

    # assert cmd_line.text() == ""
    # assert editor.textCursor().position() == 2

    qtbot.keyClicks(cmd_line, "k")

    event = QKeyEvent(QEvent.KeyPress, Qt.Key_F, Qt.ControlModifier)
    vim.vim_cmd.commandline.keyPressEvent(event)

    # assert cmd_line.text() == ""
    # assert editor.textCursor().position() == 2


def test_message(vim_bot):
    """Test message."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\nd\ne")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "y2j")
    assert vim.vim_cmd.msg_label.text() == "3 lines yanked"

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "p")
    assert vim.vim_cmd.msg_label.text() == "3 more lines"

    vim.vim_cmd.msg_label.clear()
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "vp")
    assert vim.vim_cmd.msg_label.text() == "4 more lines"

    vim.vim_cmd.msg_label.clear()
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, "Vp")
    assert vim.vim_cmd.msg_label.text() == "2 more lines"

    qtbot.keyClicks(cmd_line, "i")
    cmd_line.focusInEvent(QFocusEvent(QEvent.FocusIn, Qt.OtherFocusReason))
    assert vim.vim_cmd.msg_label.text() == ""

    qtbot.keyClicks(cmd_line, "d2j")
    assert vim.vim_cmd.msg_label.text() == "3 fewer lines"

    editor.set_text("a\nb\nc\nd\ne")
    qtbot.keyClicks(cmd_line, "c2j")
    assert vim.vim_cmd.msg_label.text() == "2 fewer lines"

    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyPress(cmd_line, Qt.Key_Enter)
    assert vim.vim_cmd.msg_label.text() == ""

    vim.vim_cmd.msg_label.setText("a")
    qtbot.keyClicks(cmd_line, "/")
    qtbot.keyPress(cmd_line, Qt.Key_Enter)
    assert vim.vim_cmd.msg_label.text() == ""

    vim.vim_cmd.msg_label.setText("a")
    qtbot.keyClicks(cmd_line, "v/")
    qtbot.keyPress(cmd_line, Qt.Key_Enter)
    assert vim.vim_cmd.msg_label.text() == ""

    vim.vim_cmd.msg_label.setText("a")
    qtbot.keyClicks(cmd_line, "V/")
    qtbot.keyPress(cmd_line, Qt.Key_Enter)
    qtbot.keyPress(cmd_line, Qt.Key_Escape)
    assert vim.vim_cmd.msg_label.text() == ""

    vim.vim_cmd.msg_label.setText("a")
    editor.set_text("a\nb\nc\nd\ne")
    qtbot.keyClicks(cmd_line, "hx")
    qtbot.keyClicks(cmd_line, "u")
    assert vim.vim_cmd.msg_label.text() == "1 changes"

    vim.vim_cmd.msg_label.setText("a")
    qtbot.keyPress(cmd_line, Qt.Key_R, Qt.ControlModifier)
    assert vim.vim_cmd.msg_label.text() == "1 changes"

    qtbot.keyClicks(cmd_line, "d3j")
    qtbot.keyClicks(cmd_line, "u")
    assert vim.vim_cmd.msg_label.text() == "4 more lines"

    qtbot.keyPress(cmd_line, Qt.Key_R, Qt.ControlModifier)
    assert vim.vim_cmd.msg_label.text() == "4 fewer lines"

    qtbot.keyClicks(cmd_line, "p")
    qtbot.keyClicks(cmd_line, "u")
    assert vim.vim_cmd.msg_label.text() == "4 fewer lines"

    qtbot.keyPress(cmd_line, Qt.Key_R, Qt.ControlModifier)
    assert vim.vim_cmd.msg_label.text() == "4 more lines"

    editor.set_text("aaaaaa")
    qtbot.keyClicks(cmd_line, "/a")
    qtbot.keyPress(cmd_line, Qt.Key_Enter)
    assert vim.vim_cmd.msg_label.text() == "/a"

    vim.vim_cmd.msg_label.setText("a")
    qtbot.keyClicks(cmd_line, "n")
    assert vim.vim_cmd.msg_label.text() == "/a"

    qtbot.keyClicks(cmd_line, "N")
    assert vim.vim_cmd.msg_label.text() == "?a"



"""Conftest."""
# Standard library imports
import os
import os.path as osp
from unittest.mock import Mock

# Third party imports
import pytest
import requests
from pytestqt.plugin import QtBot
from qtpy.QtGui import QFont
from qtpy.QtWidgets import QVBoxLayout, QWidget
from spyder.config.manager import CONF
from spyder.plugins.editor.widgets.editorstack import EditorStack

# Local imports
from spyder_okvim.spyder.config import CONF_DEFAULTS, CONF_SECTION
from spyder_okvim.spyder.plugin import OkVim

LOCATION = osp.realpath(osp.join(os.getcwd(), osp.dirname(__file__)))


class VimTesting(OkVim):
    CONF_FILE = False
    CONF_SECTION = CONF_SECTION


class EditorMock(QWidget):
    """Editor plugin mock."""

    def __init__(self, editor_stack):
        """Editor Mock constructor."""
        QWidget.__init__(self, None)
        self.editor_stack = editor_stack
        self.editorsplitter = self.editor_stack
        self.open_action = Mock()
        self.new_action = Mock()
        self.save_action = Mock()
        self.close_action = Mock()

        layout = QVBoxLayout()
        layout.addWidget(self.editor_stack)
        self.setLayout(layout)

    def get_widget(self):
        """Return self, mimicking Spyder 6 API."""
        return self

    def get_current_editorstack(self):
        """Return EditorStack instance."""
        return self.editor_stack


class StatusBarMock(QWidget):
    """Stausbar mock."""

    def __init__(self):
        QWidget.__init__(self, None)
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

    def add_status_widget(self, widget):
        self.layout.addWidget(widget)


class MainMock(QWidget):
    """Spyder MainWindow mock."""

    def __init__(self, editor_stack, qtbot_module):
        """Main Window Mock constructor."""
        QWidget.__init__(self, None)
        self.main = QWidget()
        self.plugin_focus_changed = Mock()
        self.editor = EditorMock(editor_stack)
        self.status_bar = StatusBarMock()
        self.projects = Mock()
        self.open_file = Mock()
        self.projects.get_active_project_path = lambda: None
        layout = QVBoxLayout()
        layout.addWidget(self.editor)
        layout.addWidget(self.status_bar)
        self.setLayout(layout)

        for name, dict_info in CONF_DEFAULTS:
            if name != CONF_SECTION:
                continue
            for key, val in dict_info.items():
                CONF.set(name, key, val)

        self.add_dockwidget = Mock()

        qtbot_module.add_widget(self.main)
        qtbot_module.add_widget(self.editor)
        qtbot_module.add_widget(self.status_bar)

    def get_plugin(self, dummy, error=True):
        return self.status_bar


@pytest.fixture(scope="session")
def qtbot_module(qapp, request):
    """Module fixture for qtbot."""
    result = QtBot(request)
    return result


@pytest.fixture(scope="session", autouse=True)
def vim_bot(qtbot_module):
    """Editorstack pytest fixture."""
    text = (
        "   123\n" "line 1\n" "line 2\n" "line 3\n" "line 4"
    )  # a newline is added at end
    editor_stack = EditorStack(None, [])

    # Fix the area of the selection
    font = QFont("Courier New")
    font.setPointSize(10)
    editor_stack.set_default_font(font)
    editor_stack.setMinimumWidth(400)
    editor_stack.setMinimumHeight(400)

    editor_stack.set_find_widget(Mock())
    editor_stack.set_io_actions(Mock(), Mock(), Mock(), Mock())
    finfo0 = editor_stack.new(osp.join(LOCATION, "foo.py"), "utf-8", text)
    finfo1 = editor_stack.new(osp.join(LOCATION, "foo1.py"), "utf-8", text)
    finfo2 = editor_stack.new(osp.join(LOCATION, "foo2.py"), "utf-8", text)
    finfo3 = editor_stack.new(osp.join(LOCATION, "foo3.py"), "utf-8", text)
    main = MainMock(editor_stack, qtbot_module)

    vim = VimTesting(main, None)
    vim.on_initialize()

    qtbot_module.addWidget(editor_stack)
    qtbot_module.addWidget(finfo0.editor)
    qtbot_module.addWidget(finfo1.editor)
    qtbot_module.addWidget(finfo2.editor)
    qtbot_module.addWidget(finfo3.editor)
    qtbot_module.addWidget(main)

    # Show GUI
    main.show()
    yield main, editor_stack, finfo0.editor, vim, qtbot_module

    finfo0.editor.close()
    finfo1.editor.close()
    finfo2.editor.close()
    finfo3.editor.close()
    finfo0.deleteLater()
    finfo1.deleteLater()
    finfo2.deleteLater()
    finfo3.deleteLater()
    editor_stack.deleteLater()




# -*- coding: utf-8 -*-
"""."""
# Local imports
from spyder_okvim.executor.executor_base import ExecutorSubBase


class ExecutorColon(ExecutorSubBase):
    """Execute the method by enter."""

    def __init__(self, vim_status):
        super().__init__(vim_status)
        self.allow_leaderkey = False

        self.editor_widget = vim_status.editor_widget

    def __call__(self, txt):
        """Parse txt and executor command.

        Returns
        -------
        bool
            if return is True, Clear command line

        """
        if txt[-1] == "\b":
            cmd_line = self.vim_status.cmd_line
            if len(txt) <= 2:
                self.vim_status.sub_mode = None
                return True
            else:
                cmd_line.setText(txt[:-2])
        if txt[-1] != "\r":
            return False

        if txt[0] != ":" or len(txt) <= 2:
            self.vim_status.sub_mode = None
            return True

        # ':' is saved to input cmd_info at ExecutorBase.
        # So we need only txt[1:].
        self.update_input_cmd_info(None, None, txt[1:])

        txt = txt[1:-1]  # remove :, \r
        cmd = txt.split(None, 1)
        args = cmd[1] if len(cmd) > 1 else ""
        cmd = cmd[0]

        for symbol, text in self.SYMBOLS_REPLACEMENT.items():
            cmd = cmd.replace(symbol, text)

        try:
            method = self.__getattribute__(cmd)
        except AttributeError:
            print("unknown command", cmd)
        else:
            method(args)

        self.vim_status.sub_mode = None

        return True

    def w(self, arg=""):
        """Save current file."""
        self.editor_widget.get_widget().save_action.trigger()
        self.vim_status.cursor.draw_vim_cursor()

    def q(self, arg=""):
        """Close current file."""
        self.editor_widget.get_widget().close_action.trigger()
        self.vim_status.set_focus_to_vim()

    def qexclamation(self, arg=""):
        """Close current file without saving."""
        # TODO :
        self.editor_widget.get_widget().close_action.trigger()
        self.vim_status.set_focus_to_vim()

    def wq(self, arg=""):
        """Save and close current file."""
        self.w(arg)
        self.q()

    def n(self, args=""):
        """Create new file."""
        self.editor_widget.get_widget().new_action.trigger()
        self.vim_status.set_focus_to_vim()



# -*- coding: utf-8 -*-
"""Path Finder."""
# Standard library imports
import os.path as osp
import re
import sys

# Third party imports
from qtpy.QtCore import QDir, QDirIterator, QStringListModel, Qt, Signal
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QLineEdit,
    QListView,
    QVBoxLayout,
)
from spyder.config.gui import get_font


def fuzzyfinder(query, collection):
    """Fuzzy search.

    Ref: https://github.com/amjith/fuzzyfinder

    Args:
        input (str): A partial string which is typically entered by a user.
        collection (iterable): A collection of strings which will be filtered
                               based on the `input`.

    Returns:
        suggestions (generator): A generator object that produces a list of
            suggestions narrowed down from `collection` using the `input`.
    """
    suggestions = []
    pat = ".*?".join(map(re.escape, query))
    # lookahead regex to manage overlapping matches
    pat = "(?=({0}))".format(pat)
    regex = re.compile(pat, re.IGNORECASE)

    for txt in collection:
        r = list(regex.finditer(txt))
        if r:
            best = min(r, key=lambda x: len(x.group(1)))  # find shortest match
            suggestions.append((len(best.group(1)), txt))

    return [z[-1] for z in sorted(suggestions)]


class PathFinderEdit(QLineEdit):
    """Line editor for input of path finder."""

    sig_up_key_pressed = Signal()
    sig_pg_up_key_pressed = Signal()
    sig_pg_half_up_key_pressed = Signal()
    sig_down_key_pressed = Signal()
    sig_pg_down_key_pressed = Signal()
    sig_pg_half_down_key_pressed = Signal()
    sig_enter_key_pressed = Signal()
    sig_esc_key_pressed = Signal()

    def __init__(self, parent, textChanged):
        """Init."""
        super().__init__(parent, textChanged=textChanged)
        self.dispatcher_nomodifier = {
            Qt.Key_Up: self.sig_up_key_pressed.emit,
            Qt.Key_Down: self.sig_down_key_pressed.emit,
            Qt.Key_Enter: self.sig_enter_key_pressed.emit,
            Qt.Key_Return: self.sig_enter_key_pressed.emit,
            Qt.Key_Escape: self.sig_esc_key_pressed.emit,
            Qt.Key_PageUp: self.sig_pg_up_key_pressed.emit,
            Qt.Key_PageDown: self.sig_pg_down_key_pressed.emit,
        }

        self.dispatcher_ctrl = {
            Qt.Key_P: self.sig_up_key_pressed.emit,
            Qt.Key_B: self.sig_pg_up_key_pressed.emit,
            Qt.Key_U: self.sig_pg_half_up_key_pressed.emit,
            Qt.Key_N: self.sig_down_key_pressed.emit,
            Qt.Key_F: self.sig_pg_down_key_pressed.emit,
            Qt.Key_D: self.sig_pg_half_down_key_pressed.emit,
        }

    def keyPressEvent(self, e):
        """Override Qt method."""
        key = e.key()
        modifier = e.modifiers()
        pressed_ctrl = modifier == Qt.ControlModifier
        pressed_nomodifier = modifier == Qt.NoModifier

        if pressed_nomodifier and key in self.dispatcher_nomodifier.keys():
            self.dispatcher_nomodifier[key]()
            return

        if pressed_ctrl and key in self.dispatcher_ctrl.keys():
            self.dispatcher_ctrl[key]()
            return

        super().keyPressEvent(e)


class PathFinder(QDialog):
    """Search path of files in the folder."""

    _MIN_WIDTH = 800
    _MAX_HEIGHT = 600

    def __init__(self, folder, parent=None):
        """Init."""
        super().__init__(parent)
        font = get_font()

        self.folder = folder
        self.path_selected = ""
        self.path_list = []
        self.results_old = {}
        self.setWindowTitle("Path Finder")
        self.setWindowFlags(Qt.Popup | Qt.FramelessWindowHint)
        self.setWindowOpacity(0.95)
        self.setFixedHeight(self._MAX_HEIGHT)
        self.setFont(font)

        # Set List widget
        self.list_viewer = QListView(self)
        self.list_viewer.setAttribute(Qt.WA_TransparentForMouseEvents)
        self.list_viewer.setFocusPolicy(Qt.NoFocus)
        self.list_viewer.setFixedWidth(self._MIN_WIDTH)
        self.list_viewer.setUniformItemSizes(True)
        self.list_model = QStringListModel()
        self.list_viewer.setModel(self.list_model)
        self.list_viewer.setFont(font)

        # Set edit
        self.edit = PathFinderEdit(self, textChanged=self.update_list)
        self.edit.setFont(font)
        self.edit.sig_esc_key_pressed.connect(self.close)
        self.edit.sig_enter_key_pressed.connect(self.enter)

        self.edit.sig_up_key_pressed.connect(lambda: self.prev_row(1))
        self.edit.sig_pg_up_key_pressed.connect(self.pg_up)
        self.edit.sig_pg_half_up_key_pressed.connect(self.pg_half_up)

        self.edit.sig_down_key_pressed.connect(lambda: self.next_row(1))
        self.edit.sig_pg_down_key_pressed.connect(self.pg_down)
        self.edit.sig_pg_half_down_key_pressed.connect(self.pg_half_down)

        layout = QVBoxLayout()
        layout.addWidget(self.edit)
        layout.addWidget(self.list_viewer)

        self.setLayout(layout)

        self.get_path_list()
        self.update_list()

        self.edit.setFocus()

    def get_path_selected(self):
        """Get path selected."""
        return self.path_selected

    def get_number_of_visible_lines(self):
        """Get the number of visible lines in list."""
        num_lines = 0
        lv = self.list_viewer
        height = lv.visualRect(lv.model().index(0, 0)).height()
        if height > 0:
            num_lines = lv.viewport().height() // height
        return num_lines

    def get_path_list(self):
        """Get path of files including subdiretorys."""
        if self.folder is None or not osp.isdir(self.folder):
            self.edit.setPlaceholderText("The project is not valid.")
        else:
            dir_ = QDir(self.folder)
            it = QDirIterator(
                self.folder,
                ["*.py", "*.txt", "*.md"],
                QDir.Files | QDir.NoDotAndDotDot | QDir.NoSymLinks,
                QDirIterator.Subdirectories,
            )
            while it.hasNext():
                self.path_list.append(dir_.relativeFilePath(it.next()))

        self.results_old[""] = self.path_list

    def update_list(self):
        """Update listview."""
        query = self.edit.text()
        query = query.replace(" ", "")
        paths = self.results_old.get(query, None)

        if paths is None:
            collections = self.results_old.get(query[:-1], self.path_list)
            paths = fuzzyfinder(query, collections)

        if paths:
            self.list_model.setStringList(paths)
            self.list_viewer.setCurrentIndex(self.list_model.index(0))

    def prev_row(self, stride=1):
        """Select prev row in list viewer."""
        prev_row = self.list_viewer.currentIndex().row() - stride
        prev_row = max([prev_row, 0])
        self.list_viewer.setCurrentIndex(self.list_model.index(prev_row))

    def next_row(self, stride=1):
        """Select next row in list viewer."""
        n_row = self.list_model.rowCount()
        if n_row == 0:
            return
        next_row = self.list_viewer.currentIndex().row() + stride
        next_row = min([next_row, n_row - 1])
        self.list_viewer.setCurrentIndex(self.list_model.index(next_row))

    def pg_up(self):
        """Scroll windows 1page backwards."""
        n_line = self.get_number_of_visible_lines()
        self.prev_row(n_line)

    def pg_down(self):
        """Scroll windows 1page forwards."""
        n_line = self.get_number_of_visible_lines()
        self.next_row(n_line)

    def pg_half_up(self):
        """Scroll windows half page backwards."""
        n_line = self.get_number_of_visible_lines()
        self.prev_row(n_line // 2)

    def pg_half_down(self):
        """Scroll windows half page forwards."""
        n_line = self.get_number_of_visible_lines()
        self.next_row(n_line // 2)

    def enter(self):
        """Select next row in list viewer."""
        idx = self.list_viewer.currentIndex()
        rel_path = idx.data(Qt.DisplayRole)
        if rel_path and isinstance(self.folder, str):
            path = osp.join(self.folder, rel_path)
            self.path_selected = path
        self.close()



# -*- coding: utf-8 -*-
"""Module for action of vim."""
# %% Import
# Standard library imports
import re

# Third party imports
from qtpy.QtGui import QTextCursor

# Local imports
from spyder_okvim.utils.helper_motion import HelperMotion, MotionInfo, MotionType
from spyder_okvim.utils.vim_status import VimState


class HelperAction:
    """Helper for vim action."""

    def __init__(self, vim_status):
        self.vim_status = vim_status
        self.get_editor = vim_status.get_editor
        self.get_cursor = vim_status.get_cursor
        self.set_cursor = vim_status.set_cursor
        self.set_cursor_pos = vim_status.cursor.set_cursor_pos

        self.get_block_no_start_in_selection = (
            vim_status.get_block_no_start_in_selection
        )
        self.get_block_no_end_in_selection = vim_status.get_block_no_end_in_selection
        self.get_pos_start_in_selection = vim_status.get_pos_start_in_selection
        self.get_pos_end_in_selection = vim_status.get_pos_end_in_selection
        self.helper_motion = HelperMotion(vim_status)

    def join_lines(self, cursor_pos_start: int, block_no_start: int, block_no_end: int):
        """Join lines."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        editor = self.get_editor()
        cursor = editor.textCursor()
        n_block = editor.blockCount()

        if block_no_end >= n_block - 1:
            block_no_end = n_block - 1

        if block_no_start == block_no_end:
            return

        n_line = block_no_end - block_no_start + 1
        text_list = [""]
        cursor.setPosition(cursor_pos_start)
        for _ in range(n_line - 1):
            cursor.movePosition(QTextCursor.NextBlock)
            text = cursor.block().text().lstrip()
            if text:
                text_list.append(text)

        # Replace text
        cursor.setPosition(cursor_pos_start)
        cursor.movePosition(QTextCursor.EndOfLine)
        cursor.movePosition(QTextCursor.NextBlock, QTextCursor.KeepAnchor, n_line - 1)
        cursor.movePosition(QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
        cursor.insertText(" ".join(text_list))

        # Move position of cursor
        cursor.movePosition(QTextCursor.EndOfBlock)
        cursor.movePosition(QTextCursor.Left, n=len(text_list[-1]) + 1)
        pos = cursor.position()

        self.set_cursor_pos(pos)

        editor.document_did_change()

    def replace_txt_with_ch(self, pos_start: int, pos_end: int, ch: str):
        """Replace selected text with character of argument."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        if self.vim_status.is_normal() and pos_start == pos_end:
            self.vim_status.cursor.set_cursor_pos(pos_start)
            return

        cursor = self.get_cursor()
        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText().replace("\u2029", "\n")
        text_sub = re.sub(r".", ch, text)
        cursor.insertText(text_sub)

        if self.vim_status.is_normal():
            self.vim_status.cursor.set_cursor_pos(pos_end - 1)

        editor = self.get_editor()
        editor.document_did_change()

    def _add_surrounding(self, ch: str, text: str) -> str:
        """Add surroundings."""
        prefix_dict = {
            "'": "'",
            '"': '"',
            "(": "( ",
            "{": "{ ",
            "[": "[ ",
            ")": "(",
            "}": "{",
            "]": "[",
        }

        suffix_dict = {
            "'": "'",
            '"': '"',
            "(": " )",
            "{": " }",
            "[": " ]",
            ")": ")",
            "}": "}",
            "]": "]",
        }

        return prefix_dict[ch] + text + suffix_dict[ch]

    def add_surrounding(self, pos_start: int, pos_end: int, ch: str):
        """Add surrouding."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        if self.vim_status.is_normal() and pos_start == pos_end:
            self.vim_status.cursor.set_cursor_pos(pos_start)
            return

        cursor = self.get_cursor()
        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText().replace("\u2029", "\n")
        text_sub = self._add_surrounding(ch, text)
        cursor.insertText(text_sub)

        editor = self.get_editor()
        editor.document_did_change()

    def _delete_surrounding(self, ch: str, text: str) -> str:
        """Delete surroundings."""
        open_brackets = "([{"
        text_sub = text[1:-1]
        if ch in open_brackets:
            text_sub = text_sub.strip()
        return text_sub

    def delete_surrounding(self, ch: str) -> MotionInfo:
        """Delete surrouding."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        if ch in "'\"":
            motion_info = self.helper_motion.i_quote(ch)
            if motion_info.sel_start and motion_info.sel_end:
                motion_info.sel_start -= 1
                motion_info.sel_end += 1
        else:
            motion_info = self.helper_motion.a_bracket(1, ch)

        if motion_info.sel_start is None:
            return motion_info

        cursor = self.get_cursor()
        cursor.setPosition(motion_info.sel_start)
        cursor.setPosition(motion_info.sel_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText().replace("\u2029", "\n")
        text_sub = self._delete_surrounding(ch, text)
        cursor.insertText(text_sub)

        editor = self.get_editor()
        editor.document_did_change()

        return motion_info

    def change_surrounding(self, ch_delete: str, ch_insert: str) -> MotionInfo:
        """Change surrouding."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        if ch_delete in "'\"":
            motion_info = self.helper_motion.i_quote(ch_delete)
            if motion_info.sel_start and motion_info.sel_end:
                motion_info.sel_start -= 1
                motion_info.sel_end += 1
        else:
            motion_info = self.helper_motion.a_bracket(1, ch_delete)

        if motion_info.sel_start is None:
            return motion_info

        cursor = self.get_cursor()
        cursor.setPosition(motion_info.sel_start)
        cursor.setPosition(motion_info.sel_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText().replace("\u2029", "\n")
        text_sub = self._delete_surrounding(ch_delete, text)
        text_sub = self._add_surrounding(ch_insert, text_sub)
        cursor.insertText(text_sub)

        editor = self.get_editor()
        editor.document_did_change()

        return motion_info

    def handle_case(self, motion_info: MotionInfo, method):
        """Swap case."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        if not self.vim_status.is_normal():
            sel_start = self.get_pos_start_in_selection()
            sel_end = self.get_pos_end_in_selection()
            self._handle_case(sel_start, sel_end, method)
            self.vim_status.to_normal()
            self.vim_status.cursor.set_cursor_pos(sel_start)
        elif motion_info.motion_type == MotionType.BlockWise:
            self._handle_case(motion_info.sel_start, motion_info.sel_end, method)
            self.vim_status.cursor.set_cursor_pos(motion_info.sel_start)
        else:
            cursor = self.get_cursor()
            cursor_pos_cur = cursor.position()
            cursor_pos_new = motion_info.cursor_pos
            if cursor_pos_new is None:
                return
            pos_start, pos_end = sorted([cursor_pos_cur, cursor_pos_new])

            if motion_info.motion_type == MotionType.CharWise:
                self._handle_case(pos_start, pos_end, method)
            elif motion_info.motion_type == MotionType.CharWiseIncludingEnd:
                self._handle_case(pos_start, pos_end + 1, method)
            elif motion_info.motion_type == MotionType.LineWise:
                block_start, _ = self.vim_status.cursor.get_block(pos_start)
                block_end, _ = self.vim_status.cursor.get_block(pos_end)
                sel_start = block_start.position()
                sel_end = block_end.position() + block_end.length() - 1
                self._handle_case(sel_start, sel_end, method)

                # Exception case, guu, gUU, g~~
                if pos_start == pos_end:
                    pos_start = sel_start
            self.vim_status.cursor.set_cursor_pos(pos_start)

    def _handle_case(self, pos_start: int, pos_end: int, method):
        """Swap case."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        cursor = self.get_cursor()

        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        text = cursor.selectedText()
        if method == "swap":
            cursor.insertText(text.swapcase())
        elif method == "lower":
            cursor.insertText(text.lower())
        else:
            cursor.insertText(text.upper())

        editor = self.get_editor()
        editor.document_did_change()

    def indent(self, motion_info: MotionInfo):
        """Shift lines rightwards."""
        cursor = self.get_cursor()
        cursor_pos_cur = cursor.position()
        cursor_pos_new = motion_info.cursor_pos
        if cursor_pos_new is None:
            return
        pos_start, pos_end = sorted([cursor_pos_cur, cursor_pos_new])
        self._indent(pos_start, pos_end)

    def _indent(self, pos_start: int, pos_end: int):
        """Shift lines rightwards."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        editor = self.get_editor()
        cursor = self.get_cursor()

        get_block = self.vim_status.cursor.get_block
        block_start, block_no_start = get_block(pos_start)
        block_end, block_no_end = get_block(pos_end)

        pos_start = block_start.position()
        pos_end = block_end.position() + block_end.length() - 1

        text_list = []
        for no in range(block_no_start, block_no_end + 1):
            block = editor.document().findBlockByNumber(no)
            text_list.append(block.text())

        indent = self.vim_status.indent
        text_list_indent = []
        for text in text_list:
            if text:
                text_list_indent.append(indent + text)
            else:
                text_list_indent.append("")
        texts_indent = "\n".join(text_list_indent)

        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        cursor.insertText(texts_indent)

        block_start, _ = self.vim_status.cursor.get_block(pos_start)
        len_blank = len(block_start.text()) - len(block_start.text().lstrip())

        self.vim_status.cursor.set_cursor_pos(block_start.position() + len_blank)
        editor.document_did_change()

    def unindent(self, motion_info: MotionInfo):
        """Shift lines leftwards."""
        cursor = self.get_cursor()
        cursor_pos_cur = cursor.position()
        cursor_pos_new = motion_info.cursor_pos
        if cursor_pos_new is None:
            return
        pos_start, pos_end = sorted([cursor_pos_cur, cursor_pos_new])
        self._unindent(pos_start, pos_end)

    def _unindent(self, pos_start: int, pos_end: int):
        """Shift lines leftwards."""
        self.vim_status.update_dot_cmd(connect_editor=False)

        editor = self.get_editor()
        cursor = self.get_cursor()

        get_block = self.vim_status.cursor.get_block
        block_start, block_no_start = get_block(pos_start)
        block_end, block_no_end = get_block(pos_end)

        pos_start = block_start.position()
        pos_end = block_end.position() + block_end.length() - 1

        text_list = []
        for no in range(block_no_start, block_no_end + 1):
            block = editor.document().findBlockByNumber(no)
            text_list.append(block.text())

        indent = self.vim_status.indent
        len_indent = len(indent)
        text_list_unindent = []
        for text in text_list:
            n_space = len(text) - len(text.lstrip())
            idx_discard = min([n_space, len_indent])
            text_list_unindent.append(text[idx_discard:])
        texts_unindent = "\n".join(text_list_unindent)

        cursor.setPosition(pos_start)
        cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        cursor.insertText(texts_unindent)

        block_start, _ = self.vim_status.cursor.get_block(pos_start)
        len_blank = len(block_start.text()) - len(block_start.text().lstrip())

        self.vim_status.cursor.set_cursor_pos(block_start.position() + len_blank)
        editor.document_did_change()

    def yank(self, motion_info: MotionInfo, is_explicit=False):
        """Yank text into register.

        Parameters
        ----------
        motion_info: MotionInfo
            motion_info
        is_explicit: Bool
            If this method is called explicitly, this arg is True.

        """
        if self.vim_status.is_normal():
            if motion_info.motion_type == MotionType.BlockWise:
                if motion_info.sel_start is None or motion_info.sel_end is None:
                    return None, None
            elif motion_info.cursor_pos is None:
                return None, None

        register_name = self.vim_status.get_register_name()
        register_type = VimState.NORMAL

        if not self.vim_status.is_normal():
            sel_start = self.get_pos_start_in_selection()
            sel_end = self.get_pos_end_in_selection()
            if self.vim_status.vim_state == VimState.VLINE:
                register_type = VimState.VLINE
        elif motion_info.motion_type == MotionType.BlockWise:
            sel_start = motion_info.sel_start
            sel_end = motion_info.sel_end
        else:
            cursor = self.get_cursor()
            cursor_pos_cur = cursor.position()
            cursor_pos_new = motion_info.cursor_pos
            sel_start, sel_end = sorted([cursor_pos_cur, cursor_pos_new])

            if motion_info.motion_type == MotionType.CharWiseIncludingEnd:
                sel_end += 1
            elif motion_info.motion_type == MotionType.LineWise:
                block_start, _ = self.vim_status.cursor.get_block(sel_start)
                block_end, _ = self.vim_status.cursor.get_block(sel_end)
                sel_start = block_start.position()
                sel_end = block_end.position() + block_end.length() - 1
                register_type = VimState.VLINE

        cursor = self.get_cursor()
        cursor.setPosition(sel_start)
        cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
        txt = cursor.selectedText().replace("\u2029", "\n")
        if register_type == VimState.VLINE:
            txt += "\n"

        self.vim_status.set_register(register_name, txt, register_type)
        if is_explicit is True and register_name == '"':
            self.vim_status.set_register("0", txt, register_type)

        # Set message
        doc = cursor.document()
        nb_start = doc.findBlock(sel_start).blockNumber()
        nb_end = doc.findBlock(sel_end).blockNumber()
        if nb_start != nb_end:
            self.vim_status.set_message(f"{nb_end - nb_start + 1} lines yanked")

        # highlight yank
        if is_explicit is True and self.vim_status.is_normal():
            if self.vim_status.running_dot_cmd is False:
                self.vim_status.cursor.highlight_yank(sel_start, sel_end)

        return sel_start, sel_end

    def _move_cursor_after_space(self, cursor):
        """Move cursor after white space."""
        txt = cursor.block().text()
        n_space = len(txt) - len(txt.lstrip())
        cursor.movePosition(QTextCursor.StartOfLine)
        cursor.movePosition(QTextCursor.Right, n=n_space)
        return cursor

    def paste_in_normal(self, num, is_lower):
        """Put the text before the cursor."""
        reg = self.vim_status.get_register()
        self.vim_status.update_dot_cmd(connect_editor=False, register_name=reg.name)

        editor = self.get_editor()
        n_block_old = editor.blockCount()

        if reg.content:
            content = reg.content * num
        else:
            content = ""

        cursor = self.get_cursor()
        if reg.type != VimState.VLINE:
            if not cursor.atBlockEnd() and is_lower:
                cursor.movePosition(QTextCursor.Right)
            cursor.insertText(content)
            pos_end = cursor.selectionEnd()
            self.vim_status.to_normal()
            self.set_cursor_pos(pos_end - 1)
        elif is_lower is False:
            cursor.movePosition(QTextCursor.StartOfLine)
            block_number_old = cursor.block().blockNumber()
            cursor.insertText(content)
            block = self.get_editor().document().findBlockByNumber(block_number_old)
            cursor.setPosition(block.position())
            cursor = self._move_cursor_after_space(cursor)
            self.set_cursor_pos(cursor.position())
        else:
            cursor.movePosition(QTextCursor.EndOfLine)
            cursor_pos_old = cursor.position()
            if cursor.atEnd():
                cursor.insertText("\n" + content[:-1])
            else:
                cursor.movePosition(QTextCursor.Right)
                cursor.insertText(content)
            cursor.setPosition(cursor_pos_old)
            cursor.movePosition(QTextCursor.Down)
            cursor = self._move_cursor_after_space(cursor)
            self.set_cursor_pos(cursor.position())

        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            self.vim_status.set_message(f"{n_block_new - n_block_old} more lines")

        editor = self.get_editor()
        editor.document_did_change()

    def paste_in_visual(self, num):
        """Put the text before the cursor."""
        editor = self.get_editor()
        reg = self.vim_status.get_register()
        self.vim_status.update_dot_cmd(connect_editor=False, register_name=reg.name)

        editor = self.get_editor()
        n_block_old = editor.blockCount()

        if reg.content:
            content = reg.content * num
            if reg.type == VimState.VLINE:
                content = "\n" + content
        else:
            content = ""

        cursor = self.get_cursor()
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()
        block_no_start = self.get_block_no_start_in_selection()

        cursor.setPosition(sel_start)
        cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
        cursor.insertText(content)

        if reg.type == VimState.VLINE:
            block = editor.document().findBlockByNumber(block_no_start + 1)
            cursor_pos_new = block.position()
        else:
            if "\n" in content:
                cursor_pos_new = sel_start
            else:
                cursor_pos_new = sel_start + len(content) - 1

        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            self.vim_status.set_message(f"{n_block_new - n_block_old} more lines")

        editor.document_did_change()

        self.vim_status.to_normal()
        self.vim_status.cursor.set_cursor_pos(cursor_pos_new)

    def paste_in_vline(self, num):
        """Put the text before the cursor."""
        editor = self.get_editor()
        reg = self.vim_status.get_register()
        self.vim_status.update_dot_cmd(connect_editor=False, register_name=reg.name)

        editor = self.get_editor()
        n_block_old = editor.blockCount()

        if reg.content:
            if reg.type == VimState.VLINE:
                content = reg.content * num
                content = content[:-1]
            else:
                content = reg.content
                if num > 1:
                    content = (reg.content + "\n") * num
                    content = content[:-1]
        else:
            content = ""

        cursor = self.get_cursor()
        sel_start = self.get_pos_start_in_selection()
        sel_end = self.get_pos_end_in_selection()

        cursor.setPosition(sel_start)
        cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
        cursor.insertText(content)

        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            self.vim_status.set_message(f"{n_block_new - n_block_old} more lines")

        editor.document_did_change()

        self.vim_status.to_normal()
        self.vim_status.cursor.set_cursor_pos(sel_start)

    def delete(self, motion_info: MotionInfo, is_insert, replace_txt=""):
        """Delete selected text."""
        reg = self.vim_status.get_register()
        if is_insert:
            self.vim_status.update_dot_cmd(connect_editor=True, register_name=reg.name)
        else:
            self.vim_status.update_dot_cmd(connect_editor=False, register_name=reg.name)

        if self.vim_status.is_normal():
            if motion_info.motion_type == MotionType.BlockWise:
                if motion_info.sel_start is None or motion_info.sel_end is None:
                    return
            elif motion_info.cursor_pos is None:
                return

        cursor = self.get_cursor()
        editor = self.get_editor()
        n_block_old = editor.blockCount()
        is_linewise = False

        if not self.vim_status.is_normal():
            sel_start = self.get_pos_start_in_selection()
            sel_end = self.get_pos_end_in_selection()
            if self.vim_status.vim_state == VimState.VLINE:
                is_linewise = True
        elif motion_info.motion_type == MotionType.BlockWise:
            sel_start = motion_info.sel_start
            sel_end = motion_info.sel_end
        else:
            cursor = self.get_cursor()
            cursor_pos_cur = cursor.position()
            cursor_pos_new = motion_info.cursor_pos
            sel_start, sel_end = sorted([cursor_pos_cur, cursor_pos_new])

            if motion_info.motion_type == MotionType.CharWiseIncludingEnd:
                sel_end += 1
            elif motion_info.motion_type == MotionType.LineWise:
                is_linewise = True

        if is_linewise:
            n_block = editor.blockCount()
            get_block = self.vim_status.cursor.get_block
            block_start, block_no_start = get_block(sel_start)
            block_end, block_no_end = get_block(sel_end)
            if is_insert:
                sel_start = block_start.position()
                sel_end = block_end.position() + block_end.length() - 1
            else:
                if block_no_start == 0 and block_no_end == n_block - 1:
                    sel_start = block_start.position()
                    sel_end = block_end.position() + block_end.length() - 1
                elif block_no_start == 0:
                    sel_start = block_start.position()
                    sel_end = block_end.position() + block_end.length()
                else:
                    sel_start = block_start.position() - 1
                    sel_end = block_end.position() + block_end.length() - 1

        cursor.setPosition(sel_start)
        cursor.setPosition(sel_end, QTextCursor.KeepAnchor)
        cursor.insertText(replace_txt)

        # Set message
        n_block_new = editor.blockCount()
        if n_block_new != n_block_old:
            self.vim_status.set_message(f"{n_block_old - n_block_new} fewer lines")

        # Set cursor pos in normal
        if self.vim_status.is_normal():
            cursor_pos_new = sel_start
            if is_linewise and is_insert is False:
                n_block = editor.blockCount()
                block_no = min([block_no_start, n_block - 1])
                block = editor.document().findBlockByNumber(block_no)
                txt = block.text()
                cursor_pos_new = block.position() + len(txt) - len(txt.lstrip())
                cursor.setPosition(cursor_pos_new)

            if is_insert:
                self.set_cursor_pos(cursor_pos_new + len(replace_txt))
            else:
                self.vim_status.cursor.set_cursor_pos_without_end(cursor_pos_new)

        editor.document_did_change()

    def toggle_comment(self, motion_info: MotionInfo):
        """Toggle comment."""
        self.vim_status.update_dot_cmd(connect_editor=False)
        editor = self.get_editor()
        if not self.vim_status.is_normal():
            pos_start = self.get_pos_start_in_selection()
            pos_end = self.get_pos_end_in_selection()
        elif motion_info.motion_type == MotionType.BlockWise:
            pos_start = motion_info.sel_start
            pos_end = motion_info.sel_end
            if not pos_start or not pos_end:
                return
        else:
            cursor = editor.textCursor()
            cursor_pos_cur = cursor.position()
            cursor_pos_new = motion_info.cursor_pos
            if cursor_pos_new is None:
                return
            pos_start, pos_end = sorted([cursor_pos_cur, cursor_pos_new])

        block_start, block_no_start = self.vim_status.cursor.get_block(pos_start)
        block_end, _ = self.vim_status.cursor.get_block(pos_end)
        pos_start = block_start.position()
        pos_end = block_end.position() + block_end.length() - 1

        new_cursor = editor.textCursor()
        new_cursor.setPosition(pos_start, QTextCursor.MoveAnchor)
        new_cursor.setPosition(pos_end, QTextCursor.KeepAnchor)
        editor.setTextCursor(new_cursor)
        editor.toggle_comment()

        new_cursor = editor.textCursor()
        block = editor.document().findBlockByNumber(block_no_start)
        pos_start = block.position()
        new_cursor.setPosition(pos_start, QTextCursor.MoveAnchor)
        editor.setTextCursor(new_cursor)

        self.vim_status.to_normal()
        self.vim_status.cursor.set_cursor_pos(pos_start)

        editor.document_did_change()



# -*- coding: utf-8 -*-
"""Tests for the executor_leader_key."""
# Standard library imports
from unittest.mock import Mock

# Third party imports
import pytest
from qtpy.QtCore import Qt


def test_auto_import(vim_bot):
    """Test auto_import."""
    _, _, editor, vim, qtbot = vim_bot

    editor.auto_import = Mock()
    editor.auto_import.auto_import = Mock()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "i")

    assert cmd_line.text() == ""
    assert editor.auto_import.auto_import.called


def test_toggle_breakpoint(vim_bot):
    """Test toggle breakpoint command."""
    _, editor_stack, _, vim, qtbot = vim_bot

    editor_stack.set_or_clear_breakpoint = Mock()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "b")

    assert cmd_line.text() == ""
    assert editor_stack.set_or_clear_breakpoint.called


def test_run_cell_and_advance(vim_bot):
    """Test run_cell_and_advance."""
    _, editor_stack, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.to_normal()

    cmd_line = vim.vim_cmd.commandline
    with qtbot.waitSignal(editor_stack.sig_trigger_run_action, timeout=1000) as blocker:
        qtbot.keyPress(cmd_line, Qt.Key_Space)
        qtbot.keyPress(cmd_line, Qt.Key_Enter)

    assert cmd_line.text() == ""
    assert blocker.signal_triggered
    assert blocker.args == ["run cell and advance"]


def test_run_selection(vim_bot):
    """Test run_selection."""
    _, editor_stack, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.to_normal()

    cmd_line = vim.vim_cmd.commandline
    with qtbot.waitSignal(editor_stack.sig_trigger_run_action, timeout=1000) as blocker:
        qtbot.keyPress(cmd_line, Qt.Key_Space)
        qtbot.keyClicks(cmd_line, "r")

    assert cmd_line.text() == ""
    assert blocker.signal_triggered
    assert blocker.args == ["run selection and advance"]

    qtbot.keyClicks(cmd_line, "Vj")
    with qtbot.waitSignal(editor_stack.sig_trigger_run_action, timeout=1000) as blocker:
        qtbot.keyPress(cmd_line, Qt.Key_Space)
        qtbot.keyClicks(cmd_line, "r")

    assert cmd_line.text() == ""
    assert blocker.signal_triggered
    assert blocker.args == ["run selection and advance"]


def test_formatting(vim_bot):
    """Test formatting."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text("a\nb\nc\n")
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.to_normal()

    editor.format_document_or_range = Mock()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "f")

    assert cmd_line.text() == ""
    assert editor.format_document_or_range.called

    editor.format_document_or_range = Mock()
    qtbot.keyClicks(cmd_line, "ggVj")

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "f")

    assert cmd_line.text() == ""
    assert editor.format_document_or_range.called


def test_open_swithcer(vim_bot):
    """Test open switcher."""
    main, _, _, vim, qtbot = vim_bot

    main.open_switcher = Mock()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "p")

    assert cmd_line.text() == ""
    assert main.open_switcher.called


def test_open_symbol_swithcer(vim_bot):
    """Test open symbol switcher."""
    main, _, _, vim, qtbot = vim_bot

    main.open_switcher = Mock()

    cmd_line = vim.vim_cmd.commandline
    qtbot.keyPress(cmd_line, Qt.Key_Space)
    qtbot.keyClicks(cmd_line, "s")

    assert cmd_line.text() == ""
    assert main.open_switcher.called



# -*- coding: utf-8 -*-
"""."""
"""Tests for the executor_colon."""
# Third party imports
import pytest
from qtpy.QtCore import Qt


@pytest.mark.parametrize(
    "text, cmd_list, cmd_line_expected",
    [
        ("", [":", "k", "k"], ":kk"),
        ("", [":", "k", "k", Qt.Key_Escape], ""),
    ],
)
def test_colon_cmd(vim_bot, text, cmd_list, cmd_line_expected):
    """Test colon command."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert cmd_line.text() == cmd_line_expected


@pytest.mark.parametrize(
    "text, cmd_list",
    [
        ("", [":", Qt.Key_Return]),
        ("", [":", Qt.Key_Left, "d", Qt.Key_Enter]),
    ],
)
def test_colon_corner_case_cmd(vim_bot, text, cmd_list):
    """Test colon command."""
    _, _, editor, vim, qtbot = vim_bot
    editor.set_text(text)
    vim.vim_cmd.vim_status.cursor.set_cursor_pos(0)
    vim.vim_cmd.vim_status.reset_for_test()

    cmd_line = vim.vim_cmd.commandline
    for cmd in cmd_list:
        if isinstance(cmd, str):
            qtbot.keyClicks(cmd_line, cmd)
        else:
            qtbot.keyPress(cmd_line, cmd)

    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.sub_mode is None


def test_colon_w_command(vim_bot):
    """Test :w."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "w")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.save_action.trigger.assert_called_once_with()
    main.editor.save_action.trigger.reset_mock()


def test_colon_q_command(vim_bot):
    """Test :q."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "q")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.close_action.trigger.assert_called_once_with()
    main.editor.close_action.trigger.reset_mock()


def test_colon_qexclamation_command(vim_bot):
    """Test :q!."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "q")
    qtbot.keyClicks(cmd_line, "!")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.close_action.trigger.assert_called_once_with()
    main.editor.close_action.trigger.reset_mock()


def test_colon_wq_command(vim_bot):
    """Test :wq."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "w")
    qtbot.keyClicks(cmd_line, "q")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.close_action.trigger.assert_called_once_with()
    main.editor.save_action.trigger.assert_called_once_with()
    main.editor.close_action.trigger.reset_mock()
    main.editor.save_action.trigger.reset_mock()


def test_colon_n_command(vim_bot):
    """Test :n."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "n")
    qtbot.keyPress(cmd_line, Qt.Key_Return)
    main.editor.new_action.trigger.assert_called_once_with()


def test_colon_backspace_command(vim_bot):
    """Test backspace in ex cmd."""
    main, editor_stack, editor, vim, qtbot = vim_bot
    cmd_line = vim.vim_cmd.commandline
    qtbot.keyClicks(cmd_line, ":")
    qtbot.keyClicks(cmd_line, "n")
    qtbot.keyPress(cmd_line, Qt.Key_Backspace)

    assert cmd_line.text() == ":"
    assert vim.vim_cmd.vim_status.sub_mode is not None

    qtbot.keyPress(cmd_line, Qt.Key_Backspace)

    assert cmd_line.text() == ""
    assert vim.vim_cmd.vim_status.sub_mode is None



# -*- coding: utf-8 -*-
"""Excecutor __init__."""
# %% Import
# Local imports
from .executor_normal import ExecutorNormalCmd
from .executor_visual import ExecutorVisualCmd
from .executor_vline import ExecutorVlineCmd
from .executor_leader import ExecutorLeaderKey
